#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC
from typing import List
import json
import gzip
from os.path import exists

from .workflow_properties import (
    BackendProperties,
    ErrorMitigationProperties,
    ClassicalOptimizer,
)
from ..backends.devices_core import DeviceBase, DeviceLocal
from ..problems import QUBO
from ..utilities import delete_keys_from_dict, is_valid_uuid, generate_uuid
from ..backends.qaoa_backend import (
    DEVICE_NAME_TO_OBJECT_MAPPER,
    DEVICE_ACCESS_OBJECT_MAPPER,
)


def check_compiled(func):
    def wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        if self.compiled:
            raise ValueError(
                "Cannot change properties of the object after compilation."
            )
        return result

    return wrapper


class Workflow(ABC):
    """
    Abstract class to represent an optimizer

    It's basic usage consists of

     #. Initialization
     #. Compilation
     #. Optimization

    Attributes
    ----------
    device: `DeviceBase`
        Device to be used by the optimizer
    backend_properties: `BackendProperties`
        The backend properties of the optimizer workflow. Use to set the
        backend properties such as the number of shots and the cvar values.
        For a complete list of its parameters and usage please see the method `set_backend_properties`
    classical_optimizer: `ClassicalOptimizer`
        The classical optimiser properties of the optimizer workflow.
        Use to set the classical optimiser needed for the classical optimisation part of the optimizer routine.
        For a complete list of its parameters and usage please see the method `set_classical_optimizer`
    local_simulators: `list[str]`
        A list containing the available local simulators
    cloud_provider: `list[str]`
        A list containing the available cloud providers
    compiled: `Bool`
        A boolean flag to check whether the optimizer object has been correctly compiled at least once
    id: TODO
    exp_tags: `dict`
        A dictionary containing the tags of the optimizer object. The user can set this
        value using the set_exp_tags method.
    problem: `Problem`
        The problem object that the optimizer will be optimizing.
    results: `Results`
        The results object that will contain the results of the optimization routine.
    """

    def __init__(self, device=DeviceLocal("vectorized")):
        """
        Initialize the optimizer class.

        Parameters
        ----------
        device: `DeviceBase`
            Device to be used by the optimizer. Default is using the local 'vectorized' simulator.
        """

        self.device = device
        self.backend_properties = BackendProperties()
        self.error_mitigation_properties = ErrorMitigationProperties()
        self.classical_optimizer = ClassicalOptimizer()
        self.local_simulators = list(DEVICE_NAME_TO_OBJECT_MAPPER.keys())
        self.cloud_provider = list(DEVICE_ACCESS_OBJECT_MAPPER.keys())
        self.available_error_mitigation_techniques = ["spam_twirling"]
        self.compiled = False

        # Initialize the identifier stamps, we initialize all the stamps needed to None
        self.header = {
            "atomic_id": None,  # the id of the run it is generated automatically in the compilation
            "experiment_id": generate_uuid(),  # the id of the experiment it is generated automatically here
            "project_id": None,
            "algorithm": None,  # qaoa or rqaoa
            "description": None,
            "run_by": None,
            "provider": None,
            "target": None,
            "cloud": None,
            "client": None,
            "execution_time_start": None,
            "execution_time_end": None,
        }

        # Initialize the experiment tags
        self.exp_tags = {}

        # Initialize the results and problem objects
        self.problem = None
        self.result = None

    def __setattr__(self, __name, __value):
        # check the attribute exp_tags is json serializable
        if __name == "exp_tags":
            try:
                json.dumps(__value)
            except:
                raise ValueError("The exp_tags attribute is not json serializable")

        return super().__setattr__(__name, __value)

    def set_header(
        self,
        project_id: str = None,
        description: str = None,
        run_by: str = None,
        provider: str = None,
        target: str = None,
        cloud: str = None,
        client: str = None,
        experiment_id: str = None,
    ):
        """
        Method to set the identification stamps of the optimizer object in self.header.

        Parameters
        ----------
        TODO : document the parameters
        """
        if project_id is not None:
            if not is_valid_uuid(project_id):
                raise ValueError(
                    "The project_id is not a valid uuid, example of a valid uuid: 8353185c-b175-4eda-9628-b4e58cb0e41b"
                )

        if not is_valid_uuid(project_id):
            raise ValueError(
                "The project_id is not a valid uuid, example of a valid uuid: 8353185c-b175-4eda-9628-b4e58cb0e41b"
            )

        if experiment_id is not None:
            if not is_valid_uuid(experiment_id):
                raise ValueError(
                    "The experiment_id is not a valid uuid, \
                        example of a valid uuid: 8353185c-b175-4eda-9628-b4e58cb0e41b"
                )
            else:
                self.header["experiment_id"] = experiment_id

        self.header["project_id"] = project_id
        self.header["description"] = description
        self.header["run_by"] = run_by
        self.header["provider"] = provider
        self.header["target"] = target
        self.header["cloud"] = cloud
        self.header["client"] = client

    def set_exp_tags(self, tags: dict):
        """
        Method to add tags to the experiment. Tags are stored in a dictionary
        (self.exp_tags) and can be used to identify the experiment.
        Name is a special tag that is used to identify the experiment in the results object,
        it will also be stored in the dictionary, and will overwrite any previous name.

        Parameters
        ----------
        name: `str`
            Name of the experiment. If None, the name will not be changed.
            If not None, the name will be changed to the new one.
        tags: `dict`
            Dictionary containing the tags to be added to the experiment.
            If the tag already exists, it will be overwritten.
        """

        self.exp_tags = {**self.exp_tags, **tags}

    @check_compiled
    def set_device(self, device: DeviceBase):
        """ "
        Specify the device to be used by the QAOA.

        Parameters
        ----------
        location: `str`
            Can be either local, qcs, or ibmq
        name: `str`
            The name of the device to be used, for local simulators please refer to `q.local_simulators`.
            For cloud providers please refer to the provider's naming conventions
        """
        self.device = device

    @check_compiled
    def set_backend_properties(self, **kwargs):
        """
        Set the backend properties

        Parameters
        -------------------
            device: DeviceBase
            prepend_state: [Union[QuantumCircuitBase,List[complex], np.ndarray]
                The state prepended to the circuit.
            append_state: [Union[QuantumCircuitBase,List[complex], np.ndarray]
                The state prepended to the circuit.
            init_hadamard: bool
            Whether to apply a Hadamard gate to the beginning of the
                QAOA part of the circuit.. Defaults to `True`
            n_shots: int
            Optional argument to specify the number of shots required to run QAOA computations
                on shot-based simulators and QPUs. Defaults to 100.
            cvar_alpha: float
                The value of alpha for the CVaR cost function
            noise_model: `qiskit.providers.aer.noise.NoiseModel`
                    The Qiskit noise model to be used for the simulation.
            qiskit_simulation_method: str, optional
                The method to be used for the simulation.
            qiskit_optimization_level: int, optional
                The qiskit.transpile optimization level to use. Choose from 0,1,2,3
            seed_simulator: int
                Optional argument to initialize a pseudorandom solution. Default None
            active_reset:
                #TODO
            rewiring:
                Rewiring scheme to be used for Pyquil.
                Either 'PRAGMA INITIAL_REWIRING "NAIVE"' or
                'PRAGMA INITIAL_REWIRING "PARTIAL"'. If None, defaults to NAIVE
            disable_qubit_rewiring: `bool`
                Disable automatic qubit rewiring on AWS braket backend
        """

        for key, value in kwargs.items():
            if hasattr(self.backend_properties, key):
                pass  # setattr(self.backend_properties, key, value)
            else:
                raise ValueError(
                    f"Specified argument `{value}` for `{key}` in set_backend_properties is not supported"
                )

        self.backend_properties = BackendProperties(**kwargs)
        return None

    @check_compiled
    def set_error_mitigation_properties(self, **kwargs):
        """
        Set the error mitigation properties, if any.

        Parameters
        ----------
            error_mitigation_technique: str
                The specific technique used to mitigate the errors. Only a simple state preparation and measurement twirling with bitflip averages, under the name "spam_twirling" is currently supported.
            n_batches: int
                The number of batches specifies the different negating schedules at random. Total number of shots is distributed accordingly.
            calibration_data_location: str
                The location of the json file containing calibration data. For spam twirling this is the measurement outcomes of an empty circuit under the bit-flip averaging.

        """
        for key, value in kwargs.items():
            if hasattr(self.error_mitigation_properties, key) and (
                kwargs["error_mitigation_technique"].lower()
                in self.available_error_mitigation_techniques
            ):
                pass  # setattr(self.error_mitigation, key, value)
            else:
                raise ValueError(
                    f"Specified argument `{value}` for `{key}` in set_error_mitigation_properties is not supported"
                )

        self.error_mitigation_properties = ErrorMitigationProperties(**kwargs)
        return None

    @check_compiled
    def set_classical_optimizer(self, **kwargs):
        """
        Set the parameters for the classical optimizer to be used in the optimizers workflow

        Parameters
        ----------
            method: str
                The classical optimization method. To see the list of supported optimizers, refer
                to `available_optimizers` in openqaoa/optimizers/qaoa_optimizer.py
            maxiter : Optional[int]
                Maximum number of iterations.
            maxfev : Optional[int]
                Maximum number of function evaluations.
            jac: str
                Method to compute the gradient vector. Choose from:
                    - ['finite_difference', 'param_shift', 'stoch_param_shift', 'grad_spsa']
            hess: str
                Method to compute the hessian. Choose from:
                    - ['finite_difference', 'param_shift', 'stoch_param_shift', 'grad_spsa']
            constraints: scipy.optimize.LinearConstraints, scipy.optimize.NonlinearConstraints
                Scipy-based constraints on parameters of optimization. Will be available soon
            bounds: scipy.optimize.Bounds
                Scipy-based bounds on parameters of optimization. Will be available soon
            tol : float
                Tolerance before the optimizer terminates; if `tol` is larger than
                the difference between two steps, terminate optimization.
            optimizer_options : dict
                Dictionary of optimiser-specific arguments.
                    stepsize : float
                        Step size of each gradient descent step.
                    decay : float
                        Stepsize decay parameter of RMSProp.
                    eps : float
                        Small number to prevent division by zero for RMSProp.
                    lambd : float
                        Small number to prevent singularity of QFIM matrix for Natural Gradient Descent.
            ramp_time: float
                The slope(rate) of linear ramp initialisation of QAOA parameters.
            jac_options : dict
                Dictionary that specifies gradient-computation options according to method chosen in 'jac'.
            hess_options : dict
                Dictionary that specifies Hessian-computation options according to method chosen in 'hess'.
            optimization_progress : bool
                Returns history of measurement outcomes/wavefunction if `True`. Defaults to `False`.
            cost_progress : bool
                Returns history of cost values if `True`. Defaults to `True`.
            parameter_log : bool
                Returns history of angles if `True`. Defaults to `True`.
            save_intermediate: bool
                If True, the intermediate parameters of the optimization and job ids,
                if available, are saved throughout the run. This is set to False by default.
        """
        for key, value in kwargs.items():
            if hasattr(self.classical_optimizer, key):
                pass  # setattr(self.classical_optimizer, key, value)
            else:
                raise ValueError(
                    "Specified argument is not supported by the Classical Optimizer"
                )

        self.classical_optimizer = ClassicalOptimizer(**kwargs)
        return None

    def compile(self, problem: QUBO):
        """
        Method that will make sure that the problem is in the correct form for
        the optimizer to run and generate the atomic id.
        This method should be extended by the child classes to include
        the compilation of the problem into the correct form for the optimizer to run.

        Parameters
        ----------
        problem: QUBO
            The problem to be optimized. Must be in QUBO form.
        """

        # check and set problem
        assert isinstance(problem, QUBO), "The problem must be converted into QUBO form"
        self.problem = problem

        # the atomic id is generated every time that it is compiled
        self.header["atomic_id"] = generate_uuid()

        # header is updated with the qubit number of the problem
        self.set_exp_tags({"qubit_number": self.problem.n})

    def optimize():
        raise NotImplementedError

    def _serializable_dict(
        self, complex_to_string: bool = False, intermediate_measurements: bool = True
    ):
        """
        Returns a dictionary with all values and attributes of the object that we want to
        return in `asdict` and `dump(s)` methods in a dictionary.
        The returned dictionary has two keys: header and data. The header contains all the data
        that can identify the experiment, while the data contains all the input and output data
        of the experiment (also the experiment tags).

        Parameters
        ----------
        complex_to_string: bool
            If True, converts all complex numbers to strings. This is useful for
            JSON serialization, for the `dump(s)` methods.
        intermediate_measurements: bool
            If True, intermediate measurements are included in the dump.
            If False, intermediate measurements are not included in the dump.
            Default is True.
        """

        # create the final data dictionary
        data = {}
        data["exp_tags"] = self.exp_tags.copy()
        data["input_problem"] = dict(self.problem) if self.problem is not None else None
        data["input_parameters"] = {
            "device": {
                "device_location": self.device.device_location,
                "device_name": self.device.device_name,
            },
            "backend_properties": dict(self.backend_properties),
            "classical_optimizer": dict(self.classical_optimizer),
        }
        # change the parameters that aren't serializable to strings
        for item in ["noise_model", "append_state", "prepend_state"]:
            if data["input_parameters"]["backend_properties"][item] is not None:
                data["input_parameters"]["backend_properties"][item] = str(
                    data["input_parameters"]["backend_properties"][item]
                )

        data["result"] = (
            self.result.asdict(False, complex_to_string, intermediate_measurements)
            if self.result not in [None, {}]
            else None
        )

        # create the final header dictionary
        header = self.header.copy()
        header["metadata"] = {
            **self.exp_tags.copy(),
            **(
                {
                    "problem_type": data["input_problem"]["problem_instance"][
                        "problem_type"
                    ]
                }
                if data["input_problem"] is not None
                else {}
            ),
            **(
                data["input_problem"]["metadata"].copy()
                if data["input_problem"] is not None
                else {}
            ),
            **{"n_shots": data["input_parameters"]["backend_properties"]["n_shots"]},
            **{
                prepend + key: data["input_parameters"]["classical_optimizer"][key]
                for prepend, key in zip(
                    ["optimizer_", "", ""], ["method", "jac", "hess"]
                )
                if not data["input_parameters"]["classical_optimizer"][key] is None
            },
        }

        # we return a dictionary (serializable_dict) that will have two keys: header and data
        serializable_dict = {
            # header is a dictionary containing all the data that can identify the experiment
            "header": header,
            # data is a dictionary containing all the input and output data of the experiment (also the experiment tags)
            "data": data,
        }

        return serializable_dict

    def asdict(self, exclude_keys: List[str] = [], options: dict = {}):
        """
        Returns a dictionary of the Optimizer object, where all objects are converted to dictionaries.

        Parameters
        ----------
        exclude_keys : List[str]
            A list of keys to exclude from the returned dictionary.
        options : dict
            A dictionary of options to pass to the method that creates the dictionary to dump.
                complex_to_string : bool
                    If True, converts complex numbers to strings. If False,
                    complex numbers are not converted to strings.
                intermediate_measurements : bool
                    If True, includes the intermediate measurements in the results.
                    If False, only the final measurements are included.

        Returns
        -------
        dict
        """

        options = {**{"complex_to_string": False}, **options}

        if exclude_keys == []:
            return self._serializable_dict(**options)
        else:
            return delete_keys_from_dict(
                obj=self._serializable_dict(**options), keys_to_delete=exclude_keys
            )

    def dumps(self, indent: int = 2, exclude_keys: List[str] = [], options: dict = {}):
        """
        Returns a json string of the Optimizer object.

        Parameters
        ----------
        indent : int
            The number of spaces to indent the result in the json file.
            If None, the result is not indented.
        exclude_keys : List[str]
            A list of keys to exclude from the json string.
        options : dict
            A dictionary of options to pass to the method that creates
            the dictionary to dump.
        intermediate_measurements : bool
            If True, includes the intermediate measurements in the results.
            If False, only the final measurements are included.

        Returns
        -------
        str
        """

        options = {**options, **{"complex_to_string": True}}

        if exclude_keys == []:
            return json.dumps(self._serializable_dict(**options), indent=indent)
        else:
            return json.dumps(
                delete_keys_from_dict(
                    obj=self._serializable_dict(**options), keys_to_delete=exclude_keys
                ),
                indent=indent,
            )

    def dump(
        self,
        file_name: str = "",
        file_path: str = "",
        prepend_id: bool = False,
        indent: int = 2,
        compresslevel: int = 0,
        exclude_keys: List[str] = [],
        overwrite: bool = False,
        options: dict = {},
    ):
        """
        Saves the Optimizer object as json file (if compresslevel is 0).
        If compresslevel is not 0, saves the Optimizer object as a .gz file
        (which should be decompressed before use).

        Parameters
        ----------
        file_name : str
            The name of the json file.
        file_path : str
            The path where the json file will be saved.
        prepend_id : bool
            If True, the name will have the following format: '{project_id}--{experiment_id}--{atomic_id}--{file_name}.json'.
            If False, the name will have the following format: '{file_name}.json'.
            Default is False.
        indent : int
            The number of spaces to indent the result in the json file.
            If None, the result is not indented.
        compresslevel : int
            The compression level to use. If 0, no compression is used and a json file is saved.
            If 1, the fastest compression method is used. If 9, the slowest
            but most effective compression method is used. And a .gz file is saved.
        exclude_keys : List[str]
            A list of keys that should not be included in the json file.
        overwrite : bool
            If True, overwrites the file if it already exists. If False,
            raises an error if the file already exists.
        options : dict
            A dictionary of options to pass to the method that creates the dictionary to dump.
        intermediate_measurements : bool
            If True, includes the intermediate measurements in the results.
            If False, only the final measurements are included.
        """

        options = {**options, **{"complex_to_string": True}}

        project_id = (
            self.header["project_id"]
            if not self.header["project_id"] is None
            else "None"
        )

        # get the full name
        if prepend_id is False and file_name == "":
            raise ValueError(
                "dump method missing argument: 'file_name'. Otherwise 'prepend_id' must be specified as True."
            )
        elif prepend_id is False:
            file = file_path + file_name
        elif file_name == "":
            file = (
                file_path
                + project_id
                + "--"
                + self.header["experiment_id"]
                + "--"
                + self.header["atomic_id"]
            )
        else:
            file = (
                file_path
                + project_id
                + "--"
                + self.header["experiment_id"]
                + "--"
                + self.header["atomic_id"]
                + "--"
                + file_name
            )

        # adding .json extension if not present and adding .gz extension if compresslevel is not 0 and not present
        file = file + ".json" if ".json" != file[-5:] else file
        if compresslevel != 0:
            file = file + ".gz" if ".gz" != file[-3:] else file

        # checking if the file already exists, and raising an error if it does and overwrite is False
        if overwrite is False and exists(file):
            raise FileExistsError(
                f"The file {file} already exists. Please change the name of the file or set overwrite=True."
            )

        # saving the file
        if compresslevel == 0:  # if compresslevel is 0, save as json file
            with open(file, "w") as f:
                if exclude_keys == []:
                    json.dump(self._serializable_dict(**options), f, indent=indent)
                else:
                    json.dump(
                        delete_keys_from_dict(
                            obj=self._serializable_dict(**options),
                            keys_to_delete=exclude_keys,
                        ),
                        f,
                        indent=indent,
                    )
        else:  # if compresslevel is not 0, save as .gz file (which should be decompressed before use)
            with gzip.open(file, "w", compresslevel=compresslevel) as f:
                f.write(
                    self.dumps(
                        indent=indent, exclude_keys=exclude_keys, options=options
                    ).encode("utf-8")
                )

        # print the file path and name
        if file_path == "":
            print(
                'Results saved as "{}" in the current directory.'.format(
                    file[len(file_path) :]
                )
            )
        else:
            print(
                'Results saved as "{}" in the folder "{}".'.format(
                    file[len(file_path) :], file_path
                )
            )

    @classmethod
    def from_dict(cls, dictionary: dict):
        """
        Creates an Optimizer object from a dictionary (which is the output of the asdict method)
        Parameters
        ----------
        dictionary : dict
            A dictionary with the information of the Optimizer object.
        Returns
        -------
        QAOA or RQAOA
        """

        # check if the class is correct
        algorithm = dictionary["header"]["algorithm"]
        assert (
            algorithm.lower() == cls.__name__.lower()
        ), f"The class {cls.__name__} does not match the algorithm ({algorithm}) of the dictionary."

        # create the object
        obj = cls()

        # header
        obj.header = dictionary["header"].copy()
        obj.header.pop("metadata", None)  # remove the metadata from the header

        # tags
        obj.exp_tags = dictionary["data"]["exp_tags"].copy()

        # problem
        obj.problem = (
            QUBO.from_dict(dictionary["data"]["input_problem"])
            if dictionary["data"]["input_problem"] is not None
            else None
        )

        # input parameters
        map_inputs = {
            "backend_properties": obj.set_backend_properties,
            "circuit_properties": obj.set_circuit_properties,
            "classical_optimizer": obj.set_classical_optimizer,
            "rqaoa_parameters": obj.set_rqaoa_parameters
            if algorithm == "rqaoa"
            else None,
        }
        for key, value in dictionary["data"]["input_parameters"].items():
            if key == "device":
                continue
            map_inputs[key](**value)

        # results
        if (
            "result" in dictionary["data"].keys()
            and dictionary["data"]["result"] is not None
        ):
            obj.result = obj.results_class.from_dict(
                dictionary["data"]["result"],
                **(
                    {"cost_hamiltonian": obj.problem.hamiltonian}
                    if algorithm == "qaoa"
                    else {}
                ),
            )

        # print a message when the object is loaded
        print(f"Loaded {cls.__name__} object.")
        print("The device has to be set manually using the set_device method.")
        print(
            f"Name of the device used was: {dictionary['data']['input_parameters']['device']}"
        )

        return obj

    @classmethod
    def loads(cls, string: str):
        """
        Creates an Optimizer object from a string (which is the output of the dumps method)
        Parameters
        ----------
        string : str
            A string with the information of the Optimizer object.
        Returns
        -------
        QAOA or RQAOA
        """
        return cls.from_dict(json.loads(string))

    @classmethod
    def load(cls, file_name: str, file_path: str = ""):
        """
        Creates an Optimizer object from a file (which is the output of the dump method)
        Parameters
        ----------
        file_name : str
            The name of the file.
        file_path : str
            The path of the file.
        Returns
        -------
        QAOA or RQAOA
        """
        file = file_path + file_name
        if ".gz" == file_name[-3:]:
            with gzip.open(file, "r") as f:
                return cls.loads(f.read().decode("utf-8"))
        else:
            with open(file, "r") as f:
                return cls.loads(f.read())



import unittest
import networkx as nx

from openqaoa import QAOA
from openqaoa.problems import MinimumVertexCover
from openqaoa.backends import create_device

class TestingLoggerClass(unittest.TestCase):
    def test_plot_probabilities(self):
        
        # Create the problem
        g = nx.circulant_graph(6, [1])
        vc = MinimumVertexCover(g, field=1.0, penalty=10).qubo

        # shot based simulator:
        q_shot = QAOA()
        q_shot_dev = create_device(location="local", name="qiskit.shot_simulator")
        q_shot.set_device(q_shot_dev)

        q_shot.compile(vc)
        q_shot.optimize()

        q_shot.result.plot_probabilities()


from .training_vqa import (
    ScipyOptimizer,
    CustomScipyGradientOptimizer,
    PennyLaneOptimizer,
)
from ..qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)
from ..backends.basebackend import VQABaseBackend


def available_optimizers():
    """
    Return a list of available optimizers.
    """

    optimizers = {
        "scipy": ScipyOptimizer.SCIPY_METHODS,
        "custom_scipy_gradient": CustomScipyGradientOptimizer.CUSTOM_GRADIENT_OPTIMIZERS,
        "custom_scipy_pennylane": PennyLaneOptimizer.PENNYLANE_OPTIMIZERS,
    }

    return optimizers


def get_optimizer(
    vqa_object: VQABaseBackend,
    variational_params: QAOAVariationalBaseParams,
    optimizer_dict: dict,
):
    """
    Initialise the specified optimizer class with provided method and
    optimizer-specific options

    Parameters
    ----------
    vqa_object:
        Backend object of class VQABaseBackend which contains information on the
        backend used to perform computations, and the VQA circuit.

    variational_params:
        Object of class QAOAVariationalBaseParams, which contains
        information on the circuit to be executed,
        the type of parametrisation, and the angles of the VQA circuit.

    optimizer_dict:
        Optimizer information dictionary used to construct
        the optimizer with specified options

    Returns
    -------
    optimizer:
        Optimizer object of type specified by specified method
    """
    SUPPORTED_OPTIMIZERS = {
        "scipy": ScipyOptimizer,
        "custom_scipy_gradient": CustomScipyGradientOptimizer,
        "custom_scipy_pennylane": PennyLaneOptimizer,
    }

    method = optimizer_dict["method"].lower()
    optimizers = available_optimizers()

    method_valid = False
    for opt_class, methods in optimizers.items():
        if method in methods:
            selected_class = opt_class
            method_valid = True

    assert method_valid, ValueError(
        f"Selected optimizer method '{method}' is not supported."
        f"Please choose from {available_optimizers()}"
    )

    optimizer = SUPPORTED_OPTIMIZERS[selected_class](
        vqa_object, variational_params, optimizer_dict
    )

    return optimizer



import unittest
import numpy as np
from openqaoa.problems import BinPacking


def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


def terms_list_isclose(terms_list1, terms_list2):
    """
    Check if the distance between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = (
                True
                if np.isclose(term1, term2) or np.isclose(term1, term2[::-1])
                else False
            )

    return bool


class TestBinPacking(unittest.TestCase):
    """Tests for the BinPacking class"""
    def test_binpacking_terms_weights_constant(self):
        """Test that BinPacking creates a correct QUBO from the provided weights and terms"""

        terms = [
            [2, 3],
            [4, 5],
            [0, 2],
            [0, 4],
            [0, 6],
            [2, 4],
            [2, 6],
            [4, 6],
            [1, 3],
            [1, 5],
            [1, 7],
            [3, 5],
            [3, 7],
            [5, 7],
            [0],
            [1],
            [2],
            [3],
            [4],
            [5],
            [6],
            [7],
        ]

        weights = [
            1.5,
            1.5,
            -0.75,
            -1.0,
            -1.5,
            0.5,
            0.75,
            1.0,
            -0.75,
            -1.0,
            -1.5,
            0.5,
            0.75,
            1.0,
            1.25,
            1.25,
            -0.875,
            -0.875,
            -1.1666666666666665,
            -1.1666666666666665,
            -1.75,
            -1.75,
        ]
        constant = 10.083333333333332
        weights_list = [3, 4]
        weight_capacity = 6
        binpacking_prob_qubo = BinPacking(
            weights_list, weight_capacity, simplifications=False
        ).qubo
        self.assertTrue(terms_list_equality(terms, binpacking_prob_qubo.terms))
        self.assertTrue(terms_list_isclose(weights, binpacking_prob_qubo.weights))
        self.assertTrue(np.isclose(constant, binpacking_prob_qubo.constant))

    def test_binpacking_terms_weights_constant_simplified(self):
        """Test that BinPacking creates a correct QUBO from the provided weights and terms"""

        terms = [
            [1, 2],
            [1, 3],
            [2, 3],
            [4, 5],
            [4, 6],
            [5, 6],
            [1, 4],
            [1, 7],
            [4, 7],
            [2, 5],
            [8, 2],
            [8, 5],
            [0, 3],
            [0, 6],
            [0, 9],
            [3, 6],
            [9, 3],
            [9, 6],
            [0],
            [1],
            [2],
            [3],
            [4],
            [5],
            [6],
            [7],
            [8],
            [9],
        ]

        weights = [
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            0.2,
            0.28,
            0.35,
            0.2,
            0.4,
            0.5,
            -0.4,
            -0.5,
            -1.0,
            0.2,
            0.4,
            0.5,
            0.4,
            -1.08,
            -0.96,
            -1.36,
            -1.1,
            -0.95,
            -1.45,
            -0.14,
            0.1,
            -0.9,
        ]
        constant = 9.29
        weights_list = [3, 4, 5]
        weight_capacity = 10
        binpacking_prob_qubo = BinPacking(weights_list, weight_capacity).qubo

        self.assertTrue(terms_list_equality(terms, binpacking_prob_qubo.terms))
        self.assertTrue(terms_list_isclose(weights, binpacking_prob_qubo.weights))
        self.assertTrue(np.isclose(constant, binpacking_prob_qubo.constant))

    def test_binpacking_terms_weights_constant_unbalanced(self):
        """Test that BinPacking creates a correct QUBO from the provided weights and terms
        using the unbalanced penalization encoding"""
        terms = [
            [1, 2],
            [1, 3],
            [2, 3],
            [4, 5],
            [4, 6],
            [5, 6],
            [1, 4],
            [2, 5],
            [0, 3],
            [0, 6],
            [3, 6],
            [0],
            [1],
            [2],
            [3],
            [4],
            [5],
            [6],
        ]

        weights = [
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            0.05,
            0.05,
            -0.1,
            -0.125,
            0.05,
            -0.475,
            -0.97,
            -0.91,
            -1.01,
            -0.9625,
            -0.8875,
            -1.0125,
        ]
        constant = 6.8775

        weights_list = [3, 4, 5]
        weight_capacity = 10
        binpacking_prob_qubo = BinPacking(
            weights_list, weight_capacity, method="unbalanced"
        ).qubo

        self.assertTrue(terms_list_equality(terms, binpacking_prob_qubo.terms))
        self.assertTrue(terms_list_isclose(weights, binpacking_prob_qubo.weights))
        self.assertTrue(np.isclose(constant, binpacking_prob_qubo.constant))

    def test_binpacking_terms_penalizations_terms_unbalanced(self):
        """Test that BinPacking creates a correct QUBO from the provided weights and terms
        using the unbalanced penalization encoding given the penalization terms"""
        terms = [[1, 2], [0, 2], [0], [1], [2]]

        weights = [0.5, -0.25, -0.25, -0.0625, -0.125]
        constant = 1.953125

        weights_list = [3, 4]
        weight_capacity = 8
        penalty = [1, 1, 1]
        binpacking_prob_qubo = BinPacking(
            weights_list,
            weight_capacity,
            penalty=penalty,
            method="unbalanced",
            simplifications=True,
        ).qubo

        self.assertTrue(terms_list_equality(terms, binpacking_prob_qubo.terms))
        self.assertTrue(terms_list_isclose(weights, binpacking_prob_qubo.weights))
        self.assertTrue(np.isclose(constant, binpacking_prob_qubo.constant))

    def test_binpacking_terms_penalizations_terms_slack(self):
        """Test that BinPacking creates a correct QUBO from the provided weights and terms
        using the unbalanced penalization encoding"""
        terms = [[1, 2], [1, 3], [0, 2], [0, 4], [2, 4], [0], [1], [2], [3], [4]]

        weights = [
            0.5,
            0.15625,
            -0.25,
            -0.5,
            0.25,
            -0.25,
            0.03125,
            -0.125,
            0.0390625,
            -0.25,
        ]
        constant = 2.7890625

        weights_list = [3, 4]
        weight_capacity = 8
        penalty = [1]
        binpacking_prob_qubo = BinPacking(
            weights_list,
            weight_capacity,
            penalty=penalty,
            method="slack",
            simplifications=True,
        ).qubo

        self.assertTrue(terms_list_equality(terms, binpacking_prob_qubo.terms))
        self.assertTrue(terms_list_isclose(weights, binpacking_prob_qubo.weights))
        self.assertTrue(np.isclose(constant, binpacking_prob_qubo.constant))

    def test_binpacking_random_problem(self):
        """Test Bin Packing random instance method"""

        seed = 1234
        np.random.seed(seed)
        min_weight = 1
        max_weight = 7
        n_items = 3
        weight_capacity = 15
        weights = list(np.random.randint(min_weight, max_weight, n_items))
        binpacking_manual_prob = BinPacking(weights, weight_capacity).qubo

        binpacking_random_prob = BinPacking.random_instance(
            n_items=3, seed=seed, weight_capacity=weight_capacity
        ).qubo

        self.assertTrue(
            terms_list_equality(
                binpacking_manual_prob.terms, binpacking_random_prob.terms
            )
        )
        self.assertEqual(binpacking_manual_prob.weights, binpacking_random_prob.weights)
        self.assertEqual(
            binpacking_manual_prob.constant, binpacking_random_prob.constant
        )

    def test_binpacking_classical_sol(self):
        """Test the Bin Packing random instance method classical solution"""

        seed = 1234
        np.random.seed(seed)
        binpacking_sol = BinPacking.random_instance(
            n_items=3, seed=seed
        ).classical_solution()

        sol = {
            "y_0": 1,
            "y_1": 0,
            "y_2": 0,
            "x_0_0": 1,
            "x_0_1": 0,
            "x_0_2": 0,
            "x_1_0": 1,
            "x_1_1": 0,
            "x_1_2": 0,
            "x_2_0": 1,
            "x_2_1": 0,
            "x_2_2": 0,
        }

        self.assertEqual(binpacking_sol, sol)

    def test_binpacking_plot(self):
        """Test Bin Packing random instance method"""
        from matplotlib.pyplot import Figure

        seed = 1234
        binpacking_random_prob = BinPacking.random_instance(n_items=3, seed=seed)
        sol = {
            "y_0": 1,
            "y_1": 0,
            "y_2": 0,
            "x_0_0": 1,
            "x_0_1": 0,
            "x_0_2": 0,
            "x_1_0": 1,
            "x_1_1": 0,
            "x_1_2": 0,
            "x_2_0": 1,
            "x_2_1": 0,
            "x_2_2": 0,
        }
        fig = binpacking_random_prob.plot_solution(sol)
        self.assertTrue(isinstance(fig, Figure))

    def test_binpacking_method_checking(self):
        """
        Checks if the method-checking returns the right error.
        """
        weights = [3, 5, 7]
        weight_capacity = 15
        method = "random"
        with self.assertRaises(ValueError) as e:
            BinPacking(weights, weight_capacity, method=method)
        self.assertEqual(
            f"The method '{method}' is not a valid method. Choose between 'slack' and 'unbalanced'",
            str(e.exception),
        )

    def test_binpacking_random_problem_checking(self):
        """
        Checks if the random min_weight equal to max_weight returns the right error.
        """
        min_weight = 5
        max_weight = 5
        with self.assertRaises(ValueError) as e:
            BinPacking.random_instance(min_weight=min_weight, max_weight=max_weight)
        self.assertEqual(
            f"min_weight: {min_weight} must be < max_weight:{max_weight}",
            str(e.exception),
        )

    def test_binpacking_classical_sol_checking(self):
        """
        Checks if the unfeasible classical solution returns the right error.
        """
        weights = [10, 10]
        weight_capacity = 8
        with self.assertRaises(ValueError) as e:
            BinPacking(
                weights=weights, weight_capacity=weight_capacity
            ).classical_solution()
        self.assertEqual("solution not found: integer infeasible", str(e.exception))

    def test_binpacking_input_weights(self):
        """
        Checks if the unfeasible classical solution returns the right error.
        """
        weights = [10.1, 10]
        weight_capacity = 8
        with self.assertRaises(TypeError) as e:
            BinPacking(
                weights=weights, weight_capacity=weight_capacity
            ).classical_solution()
        self.assertEqual(
            f"The weights must be integer numbers. Format {type(weights[0])} found.",
            str(e.exception),
        )

    def test_binpacking_input_weight_capacity(self):
        """
        Checks if the unfeasible classical solution returns the right error.
        """
        weights = [10, 10]
        weight_capacity = 8.1
        with self.assertRaises(TypeError) as e:
            BinPacking(
                weights=weights, weight_capacity=weight_capacity
            ).classical_solution()
        self.assertEqual(
            f"The weight_capacity must be integer. Format {type(weight_capacity)} found.",
            str(e.exception),
        )


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
import numpy as np
from typing import List, Tuple, Union
from copy import deepcopy

from .operators import Hamiltonian
from .gates import *
from .rotationangle import RotationAngle
from .gatemaplabel import GateMapLabel, GateMapType


class GateMap(ABC):
    def __init__(self, qubit_1: int):
        self.qubit_1 = qubit_1
        self.gate_label = None

    def decomposition(self, decomposition_type: str) -> List[Tuple]:
        try:
            return getattr(self, "_decomposition_" + decomposition_type)
        except Exception as e:
            print(e, "\nReturning default decomposition.")
            return getattr(self, "_decomposition_standard")

    @property
    @abstractmethod
    def _decomposition_standard(self) -> List[Tuple]:
        pass


class SWAPGateMap(GateMap):
    def __init__(self, qubit_1: int, qubit_2: int):
        super().__init__(qubit_1)
        self.qubit_2 = qubit_2
        self.gate_label = GateMapLabel(n_qubits=2, gatemap_type=GateMapType.FIXED)

    @property
    def _decomposition_standard(self) -> List[Tuple]:
        return [
            (CX, [self.qubit_1, self.qubit_2]),
            (CX, [self.qubit_2, self.qubit_1]),
            (CX, [self.qubit_1, self.qubit_2]),
        ]

    @property
    def _decomposition_standard2(self) -> List[Tuple]:
        return [
            (
                RZ,
                [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (
                RZ,
                [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            # X gate decomposition
            (RZ, [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
            (
                RX,
                [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (RZ, [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
            (
                RX,
                [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, -np.pi / 2)],
            ),
            # X gate decomposition
            (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
            (
                RX,
                [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
            (
                RX,
                [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, -np.pi / 2)],
            ),
            (
                RiSWAP,
                [
                    self.qubit_1,
                    self.qubit_2,
                    RotationAngle(lambda x: x, self.gate_label, np.pi),
                ],
            ),
            (CZ, [self.qubit_1, self.qubit_2]),
            # X gate decomposition
            (RZ, [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
            (
                RX,
                [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (RZ, [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
            (
                RX,
                [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, -np.pi / 2)],
            ),
            # X gate decomposition
            (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
            (
                RX,
                [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (RZ, [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
            (
                RX,
                [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, -np.pi / 2)],
            ),
        ]


class RotationGateMap(GateMap):
    def __init__(self, qubit_1: int):
        super().__init__(qubit_1)
        self.angle_value = None
        self.gate_label = GateMapLabel(n_qubits=1)

    @property
    def _decomposition_trivial(self) -> List[Tuple]:
        return self._decomposition_standard


class RYGateMap(RotationGateMap):
    @property
    def _decomposition_standard(self) -> List[Tuple]:
        return [
            (
                RY,
                [
                    self.qubit_1,
                    RotationAngle(lambda x: x, self.gate_label, self.angle_value),
                ],
            )
        ]


class RXGateMap(RotationGateMap):
    @property
    def _decomposition_standard(self) -> List[Tuple]:
        return [
            (
                RX,
                [
                    self.qubit_1,
                    RotationAngle(lambda x: x, self.gate_label, self.angle_value),
                ],
            )
        ]


class RZGateMap(RotationGateMap):
    @property
    def _decomposition_standard(self) -> List[Tuple]:
        return [
            (
                RZ,
                [
                    self.qubit_1,
                    RotationAngle(lambda x: x, self.gate_label, self.angle_value),
                ],
            )
        ]


class TwoQubitRotationGateMap(RotationGateMap):
    def __init__(self, qubit_1: int, qubit_2: int):
        super().__init__(qubit_1)
        self.qubit_2 = qubit_2
        self.gate_label = GateMapLabel(n_qubits=2)

    @property
    def _decomposition_trivial(self) -> List[Tuple]:
        low_level_gate = eval(type(self).__name__.strip("GateMap"))
        return [
            (
                low_level_gate,
                [
                    self.qubit_1,
                    self.qubit_2,
                    RotationAngle(lambda x: x, self.gate_label, self.angle_value),
                ],
            )
        ]


class RXXGateMap(TwoQubitRotationGateMap):
    @property
    def _decomposition_standard(self) -> List[Tuple]:
        return [
            (
                RY,
                [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (RX, [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
            (
                RY,
                [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (RX, [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
            (CX, [self.qubit_1, self.qubit_2]),
            (
                RZ,
                [
                    self.qubit_2,
                    RotationAngle(lambda x: x, self.gate_label, self.angle_value),
                ],
            ),
            (CX, [self.qubit_1, self.qubit_2]),
            (
                RY,
                [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (RX, [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
            (
                RY,
                [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (RX, [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
        ]


class RXYGateMap(TwoQubitRotationGateMap):
    @property
    def _decomposition_standard(self) -> List[Tuple]:
        raise NotImplementedError()


class RYYGateMap(TwoQubitRotationGateMap):
    @property
    def _decomposition_standard(self) -> List[Tuple]:
        return [
            (
                RX,
                [self.qubit_1, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (
                RX,
                [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (CX, [self.qubit_1, self.qubit_2]),
            (
                RZ,
                [
                    self.qubit_2,
                    RotationAngle(lambda x: x, self.gate_label, self.angle_value),
                ],
            ),
            (CX, [self.qubit_1, self.qubit_2]),
            (
                RY,
                [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, -np.pi / 2)],
            ),
            (
                RX,
                [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, -np.pi / 2)],
            ),
        ]


class RZXGateMap(TwoQubitRotationGateMap):
    @property
    def _decomposition_standard(self) -> List[Tuple]:
        return [
            (
                RY,
                [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (RX, [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
            (CX, [self.qubit_1, self.qubit_2]),
            (
                RZ,
                [
                    self.qubit_2,
                    RotationAngle(lambda x: x, self.gate_label, self.angle_value),
                ],
            ),
            (CX, [self.qubit_1, self.qubit_2]),
            (
                RY,
                [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi / 2)],
            ),
            (RX, [self.qubit_2, RotationAngle(lambda x: x, self.gate_label, np.pi)]),
        ]


class RZZGateMap(TwoQubitRotationGateMap):
    @property
    def _decomposition_standard(self) -> List[Tuple]:
        return [
            (CX, [self.qubit_1, self.qubit_2]),
            (
                RZ,
                [
                    self.qubit_2,
                    RotationAngle(lambda x: x, self.gate_label, self.angle_value),
                ],
            ),
            (CX, [self.qubit_1, self.qubit_2]),
        ]

    @property
    def _decomposition_standard2(self) -> List[Tuple]:
        return [
            (
                RZ,
                [
                    self.qubit_1,
                    RotationAngle(lambda x: x, self.gate_label, self.angle_value),
                ],
            ),
            (
                RZ,
                [
                    self.qubit_2,
                    RotationAngle(lambda x: x, self.gate_label, self.angle_value),
                ],
            ),
            (
                CPHASE,
                [
                    self.qubit_1,
                    self.qubit_2,
                    RotationAngle(lambda x: -2 * x, self.gate_label, self.angle_value),
                ],
            ),
        ]


class RYZGateMap(TwoQubitRotationGateMap):
    @property
    def _decomposition_standard(self) -> List[Tuple]:
        raise NotImplementedError()


class RiSWAPGateMap(TwoQubitRotationGateMap):
    """
    Parameterised-iSWAP gate
    """

    @property
    def _decomposition_standard(self) -> List[Tuple]:
        total_decomp = RXXGateMap.decomposition
        total_decomp.extend(RYYGateMap.decomposition)
        return total_decomp

    @property
    def _decomposition_standard2(self) -> List[Tuple]:
        return [
            (
                RiSWAP,
                [
                    self.qubit_1,
                    self.qubit_2,
                    RotationAngle(lambda x: x, self.gate_label, self.angle_value),
                ],
            )
        ]


class RotationGateMapFactory(object):
    PAULI_OPERATORS = ["X", "Y", "Z", "XX", "ZX", "ZZ", "XY", "YY", "YZ"]
    GATE_GENERATOR_GATEMAP_MAPPER = {
        term: eval(f"R{term}GateMap") for term in PAULI_OPERATORS
    }

    def rotationgatemap_list_from_hamiltonian(
        hamil_obj: Hamiltonian, gatemap_type: GateMapType = None
    ) -> List[RotationGateMap]:
        """
        Constructs a list of Rotation GateMaps from the input Hamiltonian Object.

        Parameters
        ----------
        hamil_obj: Hamiltonian
            Hamiltonian object to construct the circuit from
        gatemap_type: GateMapType
            Gatemap type constructed
        """

        pauli_terms = hamil_obj.terms
        output_gates = []

        one_qubit_count = 0
        two_qubit_count = 0

        for each_term in pauli_terms:
            if each_term.pauli_str in RotationGateMapFactory.PAULI_OPERATORS:
                pauli_str = each_term.pauli_str
                qubit_indices = each_term.qubit_indices
            elif each_term.pauli_str[::-1] in RotationGateMapFactory.PAULI_OPERATORS:
                pauli_str = each_term.pauli_str[::-1]
                qubit_indices = each_term.qubit_indices[::-1]
            else:
                raise ValueError("Hamiltonian contains non-Pauli terms")

            try:
                gate_class = RotationGateMapFactory.GATE_GENERATOR_GATEMAP_MAPPER[
                    pauli_str
                ]
            except Exception:
                raise Exception("Generating gates from Hamiltonian terms failed")

            if len(each_term.qubit_indices) == 2:
                gate = gate_class(qubit_indices[0], qubit_indices[1])
                gate.gate_label.update_gatelabel(
                    new_application_sequence=two_qubit_count,
                    new_gatemap_type=gatemap_type,
                )
                output_gates.append(gate)
                two_qubit_count += 1
            elif len(each_term.qubit_indices) == 1:
                gate = gate_class(qubit_indices[0])
                gate.gate_label.update_gatelabel(
                    new_application_sequence=one_qubit_count,
                    new_gatemap_type=gatemap_type,
                )
                output_gates.append(gate)
                one_qubit_count += 1

        return output_gates

    def gatemaps_layer_relabel(
        gatemap_list: List[GateMap], new_layer_number: int
    ) -> List[RotationGateMap]:
        """
        Reconstruct a new gatemap list from a list of RotationGateMap Objects with the input
        layer number in the gate_label attribute.

        Parameters
        ----------
        gatemap_list: `List[RotationGateMap]
            The list of GateMap objects whose labels need to be udpated
        """
        output_gate_list = []

        for each_gatemap in gatemap_list:
            new_gatemap = deepcopy(each_gatemap)
            new_gatemap.gate_label.update_gatelabel(new_layer_number=new_layer_number)
            output_gate_list.append(new_gatemap)

        return output_gate_list



import unittest
import pytest

from openqaoa_braket.backends import DeviceAWS


class TestingDeviceAWS(unittest.TestCase):
    """These tests check the Object used to access AWS Braket and their
    available QPUs can be established.

    For any tests using provided credentials, the tests will only pass if those
    details provided are correct/valid with AWS Braket.
    """
    
    @pytest.mark.braket_api
    def test_changing_aws_region(self):
        
        device_obj = DeviceAWS(
          device_name='arn:aws:braket:::device/quantum-simulator/amazon/sv1',
          aws_region='us-east-1'
        )
        
        device_obj.check_connection()
        default_region = device_obj.aws_region

        self.assertEqual("us-east-1", default_region)

        device_obj = DeviceAWS(
            device_name="arn:aws:braket:::device/quantum-simulator/amazon/sv1",
            aws_region="us-west-1",
        )

        device_obj.check_connection()
        custom_region = device_obj.aws_region
        
        self.assertEqual('us-west-1', custom_region)
        
    @pytest.mark.braket_api
    def test_changing_s3_bucket_names(self):

        device_obj = DeviceAWS(
            device_name="arn:aws:braket:::device/quantum-simulator/amazon/sv1",
            s3_bucket_name="random_new_name",
        )

        device_obj.check_connection()
        custom_bucket = device_obj.s3_bucket_name
        
        self.assertEqual('random_new_name', custom_bucket)
        
    @pytest.mark.braket_api      
    def test_check_connection_provider_no_backend_provided_credentials(self):

        """
        If no information about the device name, but the credentials
        used are correct, check_connection should return True.
        The provider_connected attribute should be updated to True.
        """

        device_obj = DeviceAWS(device_name="")

        self.assertEqual(device_obj.check_connection(), True)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, None)


    @pytest.mark.braket_api
    def test_check_connection_provider_right_backend_provided_credentials(self):

        """
        If the correct device name is provided and the credentials
        used are correct, check_connection should return True.
        The provider_connected attribute should be updated to True.
        The qpu_connected attribute should be updated to True.
        """

        device_obj = DeviceAWS(device_name="")

        device_obj.check_connection()
        valid_qpu_name = device_obj.available_qpus[0]

        device_obj = DeviceAWS(device_name=valid_qpu_name)

        self.assertEqual(device_obj.check_connection(), True)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, True)


    @pytest.mark.braket_api
    def test_check_connection_provider_wrong_backend_provided_credentials(self):

        """
        If device name provided is incorrect, and not empty, and the credentials
        used are correct, check_connection should return False.
        The provider_connected attribute should be updated to True.
        The qpu_connected attribute should be updated to False.
        """

        device_obj = DeviceAWS(device_name="random_invalid_backend")

        self.assertEqual(device_obj.check_connection(), False)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, False)


if __name__ == "__main__":
    unittest.main()



import numpy as np

from ..utilities import check_kwargs
from .problem import Problem
from .qubo import QUBO


class NumberPartition(Problem):
    """
    Creates an instance of the Number Partitioning problem.

    Parameters
    ----------
    numbers: List[int]
        The list of numbers to be partitioned.

    Returns
    -------
        An instance of the Number Partitioning problem.
    """

    __name__ = "number_partition"

    def __init__(self, numbers=None):
        # Set the numbers to be partitioned. If not given, generate a random list with integers
        self.numbers = numbers
        self.n_numbers = None if numbers == None else len(self.numbers)

    @property
    def numbers(self):
        return self._numbers

    @numbers.setter
    def numbers(self, input_numbers):
        if not isinstance(input_numbers, list):
            raise TypeError("The input parameter, numbers, has to be a list")

        for each_entry in input_numbers:
            if not isinstance(each_entry, int):
                raise TypeError("The elements in numbers list must be of type int.")

        self._numbers = input_numbers

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Number Partitioning problem.

        Parameters
        ----------
        n_numbers: int
            The number of numbers to be partitioned. This is a required keyword argument.

        Returns
        -------
            A random instance of the Number Partitioning problem.
        """
        n_numbers = check_kwargs(["n_numbers"], [None], **kwargs)

        # Set a random number generator
        seed = kwargs.get("seed", None)
        seed = seed if isinstance(seed, int) else None
        rng = np.random.default_rng(seed)

        numbers = list(map(int, rng.integers(1, 10, size=n_numbers)))
        return NumberPartition(numbers)

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        terms = []
        weights = []
        constant_term = 0

        # Consider every pair of numbers (ordered)
        for i in range(self.n_numbers):
            for j in range(i, self.n_numbers):
                # If i equals j, then whatever random sign we choose, if we square
                # it we can back 1. So we have a constant term.
                if i == j:
                    constant_term += self.numbers[i] * self.numbers[j]

                # Otherwise the weight is computed as being the product of the
                # numbers in the pair, multiplied by 2 (since we account for
                # both pair (i, j) and (j, i)
                else:
                    term = [i, j]
                    weight = 2 * self.numbers[i] * self.numbers[j]

                    terms.append(term)
                    weights.append(weight)

        # If the constant term is non-zero, we may add it to terms and weights
        if constant_term > 0:
            terms.append([])
            weights.append(constant_term)

        return QUBO(self.n_numbers, terms, weights, self.problem_instance)



"""
Utility and convenience functions for a number of QAOA applications.
"""

from __future__ import annotations
from typing import Optional, Union, List, Tuple
import itertools
import numpy as np
import uuid
import matplotlib.pyplot as plt
import networkx as nx
import datetime

from .qaoa_components import Hamiltonian, PauliOp

from .qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)  # ff
from .qaoa_components.ansatz_constructor.gatemap import TwoQubitRotationGateMap  # ff


def X_mixer_hamiltonian(n_qubits: int, coeffs: List[float] = None) -> Hamiltonian:
    """Construct a Hamiltonian object to implement the X mixer.

    Parameters
    ----------
    n_qubits: `int`
        The number of qubits in the mixer Hamiltonian.
    coeffs: `List[float]`
        The coefficients of the X terms in the Hamiltonian.

    Returns
    -------
    `Hamiltonian`
        The Hamiltonian object corresponding to the X mixer.
    """
    # If no coefficients provided, set all to -1
    coeffs = [-1] * n_qubits if coeffs is None else coeffs

    # Initialize list of terms
    terms = []

    # Generate terms in the X mixer
    for i in range(n_qubits):
        terms.append(PauliOp.X(i))

    # Define mixer Hamiltonian
    hamiltonian = Hamiltonian(pauli_terms=terms, coeffs=coeffs, constant=0)

    return hamiltonian


def XY_mixer_hamiltonian(
    n_qubits: int,
    qubit_connectivity: Union[List[list], List[tuple], str] = "full",
    coeffs: List[float] = None,
) -> Hamiltonian:
    r"""
    Construct a Hamiltonian object to implement the XY mixer.

    .. Important::
            The XY mixer is not implemented with :math:`RXY` Gates, but with :math:`H_{XY}=\frac{1}{2}(\sum_{i,j} X_iX_j+Y_iY_j)`

    Parameters
    ----------
    n_qubits: `int`
        The number of qubits in the system.
    qubit_connectivity: `Union[List[list],List[tuple], str]`
        The connectivity of the qubits in the mixer Hamiltonian.
    coeffs: `List[float]`
        The coefficients of the XY terms in the Hamiltonian.

    Returns
    -------
    `Hamiltonian`
        The Hamiltonian object corresponding to the XY mixer.
    """
    # Set of topologies supported by default
    connectivity_topology_dict = {
        "full": list(itertools.combinations(range(n_qubits), 2)),
        "chain": [(i, i + 1) for i in range(n_qubits - 1)],
        "star": [(0, i + 1) for i in range(n_qubits - 1)],
    }

    # Check if input connectivity is a default value
    if isinstance(qubit_connectivity, str):
        try:
            # Define qubit connectivity from default value
            qubit_connectivity = connectivity_topology_dict[qubit_connectivity]
        except KeyError:
            raise ValueError(
                f"Please choose connection topology from {list(connectivity_topology_dict.keys())}"
            )

    # Define connectivty according to user input
    else:
        # Extract indices from connectivity
        indices = set([qubit for term in qubit_connectivity for qubit in term])

        # Ensure all indices are defined within the range of number of qubits
        assert (
            max(indices) <= n_qubits - 1
        ), "Qubit index in connectivity list is out of range"
        assert min(indices) >= 0, "Qubit index should be a positive integer"

    # If no coefficients provided, set all to the number of terms
    coeffs = [0.5] * 2 * len(qubit_connectivity) if coeffs is None else coeffs

    # Initialize list of terms
    terms = []

    # Generate terms in the XY mixer
    for pair in qubit_connectivity:
        i, j = pair
        terms.append(PauliOp.X(i) @ PauliOp.X(j))
        terms.append(PauliOp.Y(i) @ PauliOp.Y(j))

    # Define mixer Hamiltonian
    hamiltonian = Hamiltonian(pauli_terms=terms, coeffs=coeffs, constant=0)

    return hamiltonian


def quick_create_mixer_for_topology(
    input_gatemap: TwoQubitRotationGateMap,
    n_qubits: int,
    qubit_connectivity: Union[List[list], List[tuple], str] = "full",
    coeffs: List[float] = None,
) -> Tuple[List[TwoQubitRotationGateMap], List[float]]:
    """
    Quickly generates a gatemap list and coeffs for a specific topology.
    Can only be used with 2-Qubit Gates.

    Parameters
    ----------
    input_gatemap: `TwoQubitRotationGateMap`
        The GateMap whose connectivity we are trying to create.
    n_qubits: `int`
        The number of qubits in the system.
    qubit_connectivity: `Union[List[list],List[tuple], str]`
        The connectivity of the qubits in the mixer.
    coeffs: `List[float]`, optional
        The coefficients of the GateMap in the Mixer Blocks.

    Returns
    -------
    `Tuple[List[TwoQubitRotationGateMap], List[float]]`
        Returns tuple containing the list of gatemaps and their associated coefficients.
        If no coefficients were on initialisation provided,
        a default of 1.0 is used for all gatemap objects.
    """

    # Set of topologies supported by default
    connectivity_topology_dict = {
        "full": list(itertools.combinations(range(n_qubits), 2)),
        "chain": [(i, i + 1) for i in range(n_qubits - 1)],
        "star": [(0, i + 1) for i in range(n_qubits - 1)],
    }

    # Check if input connectivity is a default value
    if isinstance(qubit_connectivity, str):
        try:
            # Define qubit connectivity from default value
            qubit_connectivity = connectivity_topology_dict[qubit_connectivity]
        except KeyError:
            raise ValueError(
                f"Please choose connection topology from {list(connectivity_topology_dict.keys())}"
            )

    # Define connectivty according to user input
    else:
        # Extract indices from connectivity
        indices = set([qubit for term in qubit_connectivity for qubit in term])

        # Ensure all indices are defined within the range of number of qubits
        assert (
            max(indices) <= n_qubits - 1
        ), "Qubit index in connectivity list is out of range"
        assert min(indices) >= 0, "Qubit index should be a positive integer"

    # If no coefficients provided, set all to the number of terms
    coeffs = [1.0] * len(qubit_connectivity) if coeffs is None else coeffs

    # Initialize list of terms
    gatemaps = []

    # Generate terms in the 2Q mixer
    for pair in qubit_connectivity:
        i, j = pair
        gatemaps.append(input_gatemap(i, j))

    return gatemaps, coeffs


def get_mixer_hamiltonian(
    n_qubits: int,
    mixer_type: str = "x",
    qubit_connectivity: Union[List[list], List[tuple], str] = None,
    coeffs: List[float] = None,
) -> Hamiltonian:
    """
    Parameters
    ----------
    n_qubits: `int`
        Number of qubits in the Hamiltonian.
    mixer_type: `str`
        Name of the mixer Hamiltonian. Choose from `x` or `xy`.
    qubit_connectivity: `Union[List[list],List[tuple], str]`, optional
        The connectivity of the qubits in the mixer Hamiltonian.
    coeffs: `List[float]`, optional
        The coefficients of the terms in the Hamiltonian.

    Returns:
    --------
    `Hamiltonian`
        Hamiltonian object containing the specificied mixer.
    """

    # Return mixer Hamiltonian according to specified type
    if mixer_type == "x":
        mixer = X_mixer_hamiltonian(n_qubits, coeffs)
    else:
        mixer = XY_mixer_hamiltonian(n_qubits, qubit_connectivity, coeffs)

    return mixer


################################################################################
# decorators
################################################################################
def round_value(function):
    """
    Round a value to a given precision.
    This function will be used as a decorator to round the values given by the
    ``expectation`` and ``expectation_w_uncertainty`` methods.

    Parameters
    ----------
    function: `Callable`
        The function to be decorated

    Returns
    -------
        The rounded value(s)

    """

    PRECISION = 12

    def wrapper(*args, **kwargs):
        values = function(*args, **kwargs)
        if isinstance(values, dict):
            return {k: round(v, PRECISION) for k, v in values.items()}
        else:
            return np.round(values, PRECISION)

    return wrapper


################################################################################
# METHODS FOR PRINTING HAMILTONIANS AND GRAPHS, AND PRINTING ONE FROM EACH OTHER
################################################################################


def graph_from_hamiltonian(hamiltonian: Hamiltonian) -> nx.Graph:
    """
    Creates a networkx graph corresponding to a specified problem Hamiltonian.

    .. Important::
        This function cannot handle non-QUBO terms.
        Linear terms are stored as nodes with weights.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        The Hamiltonian of interest. Must be specified a Hamiltonian object.

    Returns
    -------
    G: `networkx.Graph`
        The corresponding networkx graph with the edge weights being the
        two-qubit coupling coefficients,
        and the node weights being the single-qubit bias terms.



    """
    # Define graph
    G = nx.Graph()

    # Add nodes for each qubit in the register
    for qubit in hamiltonian.qureg:
        G.add_node(qubit, weight=0)

    # Add each term to the graph as an attribute
    for term, weight in zip(hamiltonian.terms, hamiltonian.coeffs):
        # Extract indices from Pauli term
        term_tuple = term.qubit_indices

        # If term is linear add as a node with a weight attribute
        if len(term) == 1:
            G.add_node(term_tuple[0], weight=weight)

        # If term is quadratic add as an edge with a weight attribute
        elif len(term) == 2:
            G.add_edge(term_tuple[0], term_tuple[1], weight=weight)

    return G


def hamiltonian_from_graph(G: nx.Graph) -> Hamiltonian:
    """
    Builds a cost Hamiltonian as a collection of PauliOp objects
    from a specified networkx graph, extracting any node and edge weights.

    Parameters
    ----------
    G: `networkx.Graph`
        The specified networkx graph.

    Returns
    -------
    hamiltonian: `Hamiltonian`
        The Hamiltonian object constructed from the specified graph.
    """
    # Node bias terms
    nodes_info = dict(G.nodes(data="weight"))
    singlet_terms = [
        (node,) for node, weight in nodes_info.items() if weight is not None
    ]
    singlet_coeffs = [coeff for coeff in nodes_info.values() if coeff is not None]

    # Edge terms
    pair_terms, pair_coeffs = [], []
    for u, v, edge_weight in G.edges(data="weight"):
        pair_terms.append((u, v))
        # We expect the edge weight to be given in the attribute called
        # "weight". If it is None, assume a weight of 1.0
        pair_coeffs.append(edge_weight if edge_weight else 1)

    # Collect all terms and coefficients
    terms = singlet_terms + pair_terms
    coeffs = singlet_coeffs + pair_coeffs

    # Define Hamiltonian
    hamiltonian = Hamiltonian.classical_hamiltonian(
        terms=terms, coeffs=coeffs, constant=0
    )

    return hamiltonian


def random_k_regular_graph(
    degree: int,
    nodes: List[int],
    seed: int = None,
    weighted: bool = False,
    biases: bool = False,
) -> nx.Graph:
    """
    Produces a random graph with specified number of nodes, each having degree k.

    Parameters
    ----------
    degree: `int`
        Desired degree for the nodes.
    nodes: `list`
        The node set of the graph.
    seed: `int`, optional
        A seed for the random number generator.
    weighted: `bool`, optional
        Whether the edge weights should be uniform or different. If false, all weights are set to 1.
        If true, the weight is set to a random number drawn from the uniform distribution in the
        interval 0 to 1.
    biases: `bool`, optional
        Whether or not the graph nodes should be assigned a weight.
        If true, the weight is set to a random number drawn from the uniform
        distribution in the interval 0 to 1.

    Returns
    -------
    nx.Graph: `Networkx Graph`
        A graph with the properties as specified.
    """
    # Set numpy seed
    np.random.seed(seed=seed)

    # Create a random regular graph on the nodes
    G = nx.random_regular_graph(degree, len(nodes), seed)

    # Relabel nodes
    nx.relabel_nodes(G, {i: n for i, n in enumerate(nodes)})

    # Add edges between nodes
    for edge in G.edges():
        # If weighted attribute is False, all weights are set to 1
        if not weighted:
            G[edge[0]][edge[1]]["weight"] = 1

        # If weighted attribute is True, weights are assigned as random integers
        else:
            G[edge[0]][edge[1]]["weight"] = np.random.rand()

    # If biases attribute is True, add node weights as random integers
    if biases:
        for node in G.nodes():
            G.nodes[node]["weight"] = np.random.rand()

    return G


def plot_graph(G: nx.Graph, ax=None, colormap="seismic") -> None:
    """
    Plots a networkx graph.

    Parameters
    ----------
    G: `Networkx Graph`
        The networkx graph of interest.
    ax: `Matplotlib axes object`, optional
        Matplotlib axes to plot on. Defaults to None.
    colormap: `str`, optional
        Colormap to use for plotting. Defaults to 'seismic'.
    """

    # Create plot figure
    fig = plt.figure(figsize=(10, 6))

    # Extract all graph attributes
    biases_and_nodes = nx.get_node_attributes(G, "weight")
    biases = list(biases_and_nodes.values())
    edges_and_weights = nx.get_edge_attributes(G, "weight")
    pos = nx.shell_layout(G)

    # extract minimum and maximum weights for side bar limits
    weights = list(edges_and_weights.values())
    # Define color map
    cmap = plt.cm.get_cmap(colormap)

    if len(set(weights)) > 1:
        edge_vmin = min(weights)
        edge_vmax = max(weights)

        # Define normalized color map
        sm = plt.cm.ScalarMappable(
            cmap=cmap, norm=plt.Normalize(vmin=edge_vmin, vmax=edge_vmax)
        )
        # Add colormap to plot
        cbar = plt.colorbar(sm, pad=0.08)
        cbar.ax.set_ylabel("Edge Weights", rotation=270, labelpad=15)
    else:
        weights = [1] * len(G.edges())
        edge_vmin = None
        edge_vmax = None
        cmap = None

    # If biases are present define reference values and color map for side bar
    if len(set(biases)) > 1:
        cmap = plt.cm.get_cmap(colormap)
        vmin = min(biases)
        vmax = max(biases)
        sm2 = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=vmin, vmax=vmax))
        cbar2 = plt.colorbar(sm2, location="left")
        cbar2.ax.set_ylabel("Single Qubit Biases", rotation=90)

        # Draw graph
        nx.draw(
            G,
            pos,
            node_size=500,
            node_color=biases,
            edge_color=weights,
            width=2.5,
            cmap=cmap,
            edge_cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            edge_vmin=edge_vmin,
            edge_vmax=edge_vmax,
            with_labels=True,
        )

    else:
        # Draw graph
        nx.draw(
            G,
            pos,
            node_size=500,
            edge_color=weights,
            width=2.5,
            edge_cmap=cmap,
            edge_vmin=edge_vmin,
            edge_vmax=edge_vmax,
            with_labels=True,
        )

    # Show plot

    plt.show
    return None


def random_classical_hamiltonian(
    reg: List[int],
    seed: int = None,
    weighted: bool = True,
    biases: bool = True,
    constant: int = 0,
) -> Hamiltonian:
    """
    Creates a random classical cost hamiltonian.

    Parameters
    ----------
    reg: `list`
        Register to build the hamiltonian on.
    seed: `int`, optional
        A seed for the random number generator. Defaults to None.
    weighted: `bool`, optional
        Whether the edge weights should be uniform or different. If false, all
        weights are set to 1. If true, the weight is set to a random number
        drawn from the uniform distribution in the interval 0 to 1. Defaults to
        True.
    biases: `bool`, optional
        Whether or not the graph nodes should be assigned a weight.
        If true, the weight is set to a random number drawn from the uniform
        distribution in the interval 0 to 1. Defaults to True.
    constant: `int`, optional
        The constant term in the Hamiltonian. Defaults to 0.

    Returns
    -------
    random_hamil: `Hamiltonian`
        A random hamiltonian with randomly selected terms and coefficients and
        with the specified constant term.

    .. Important::
        Randomly selects which qubits that will have a bias term, then assigns them a bias coefficient.
        Randomly selects which qubit pairs will have a coupling term, then assigns them a coupling coefficient.
        In both cases, the random coefficient is drawn from the uniform distribution on the interval [0,1).
    """
    # Set the random seed
    np.random.seed(seed=seed)

    # Initialize terms and weights
    terms = []
    weights = []

    # If biases attribute is True, add lineat terms
    if biases:
        # Choose a random set of qubits to add linear terms
        n_biases = np.random.randint(len(reg))
        bias_qubits = np.random.choice(reg, n_biases)

        # Generate coefficients for linear terms
        bias_coeffs = np.random.rand(n_biases) if weighted else np.ones(n_biases)

        # Store linear terms and coefficients
        for qubit, coeff in zip(bias_qubits, bias_coeffs):
            terms.append([qubit])
            weights.append(coeff)

    # Generate quiadratic terms, scanning all possible combinations
    for q1, q2 in itertools.combinations(reg, 2):
        # Choose at random to couple terms
        are_coupled = np.random.randint(2)

        # If coupled, generate coefficients and store along with term
        if are_coupled:
            couple_coeff = np.random.rand() if weighted else 1
            terms.append([q1, q2])
            weights.append(couple_coeff)

    # Ensure each term has an associated weight
    assert len(terms) == len(weights), "Each term should have an associated weight"

    # Define classical Hamiltonian
    hamiltonian = Hamiltonian.classical_hamiltonian(terms, weights, constant=constant)

    return hamiltonian


################################################################################
# HAMILTONIANS AND DATA
################################################################################


def ground_state_hamiltonian(
    hamiltonian: Hamiltonian, bounded=True
) -> Tuple[float, list]:
    """
    Computes the exact ground state and ground state energy of a classical Hamiltonian.
    Uses standard numpy module.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object whose ground state properties are computed.
    bounded: `bool`, optional
        If set to True, the function will not perform computations for qubit
        numbers above 25. If False, the user can specify any number. Defaults
        to True.

    Returns
    -------
    min_energy: `float`
        The minimum eigenvalue of the cost Hamiltonian.

    config: `np.array`
        The minimum energy eigenvector as a binary array
        configuration: qubit-0 as the first element in the sequence.
    """
    # Extract number of qubits
    n_qubits = hamiltonian.n_qubits

    # If number of qubits is too high warn the user
    if bounded and n_qubits > 25:
        raise ValueError(
            "The number of qubits is too high, computation could take a long time. If still want to proceed set argument `bounded` to False"
        )

    # Generate qubit register
    register = range(n_qubits)

    # Intialize energies
    energies = np.zeros(2 ** len(register))

    # Obtain spectrum, scanning term by term
    for i, term in enumerate(hamiltonian.terms):
        # Extract coefficient
        out = np.real(hamiltonian.coeffs[i])

        # Compute tensor product
        for qubit in register:
            if qubit in term.qubit_indices:
                out = np.kron([1, -1], out)
            else:
                out = np.kron([1, 1], out)

        # Add energy contribution of the term
        energies += out

    # Add constant term to the spectrum
    energies += hamiltonian.constant

    # Extract minimum energy
    min_energy = np.min(energies)

    # Extract indices of minimum energies
    indices = np.where(energies == min_energy)[0]

    # Generate ground states
    config_strings = [np.binary_repr(index, len(register))[::-1] for index in indices]

    return min_energy, config_strings


def bitstring_energy(
    hamiltonian: Hamiltonian, bitstring: Union[List[int], str]
) -> float:
    """
    Computes the energy of a given bitstring with respect to a classical cost Hamiltonian.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object determining the energy levels.
    bitstring : `list` or `str`
        A list of integers 0 and 1, or a string, representing a configuration.

    Returns
    -------
    energy: `float`
        The energy of the given bitstring with respect to the cost Hamiltonian.
    """
    # Initialize energy value
    energy = 0

    # Compute energy contribution term by term
    for i, term in enumerate(hamiltonian.terms):
        # Compute sign of spin interaction term
        variables_product = np.prod(
            [(-1) ** int(bitstring[k]) for k in term.qubit_indices]
        )

        # Add energy contribution
        energy += hamiltonian.coeffs[i] * variables_product

    # Add constant contribution
    energy += hamiltonian.constant

    return energy


def energy_expectation(hamiltonian: Hamiltonian, measurement_counts: dict) -> float:
    """
    Computes the energy expectation value from a set of measurement counts,
    with respect to a classical cost Hamiltonian.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object determining the energy levels.
    measurement_counts : `dict`
        Measurement counts dictionary for which to calculate energy expectation.

    Returns
    -------
    energy: `float`
        The energy expectation value for the set of measurement outcomes
    """

    # Starting value for the energy
    energy = 0

    # Number of measurement shots
    shots = sum(measurement_counts.values())

    # Compute average energy adding one by one the contribution from each state
    for state, prob in measurement_counts.items():
        # Number of ones (spins pointing down) from the specific
        # configuration for each Hamiltonian term
        num_ones_list = [
            sum([int(state[i]) for i in term.qubit_indices])
            for term in hamiltonian.terms
        ]

        # Compute configuration energy
        config_energy = sum(
            [
                hamiltonian.coeffs[i]
                if num_ones % 2 == 0
                else -1 * hamiltonian.coeffs[i]
                for i, num_ones in enumerate(num_ones_list)
            ]
        )

        # Add contribution to total energy
        energy += prob * config_energy

    # Normalize with respect to the number of shots
    energy *= 1 / shots

    # Add constant term in Hamiltonian
    energy += hamiltonian.constant

    return energy


def energy_spectrum_hamiltonian(hamiltonian: Hamiltonian) -> np.ndarray:
    """
    Computes exactly the energy spectrum of the hamiltonian defined by terms
    and weights and its corresponding configuration of variables. Uses
    standard numpy module.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object whose spectrum is computed.

    Returns
    -------
    energies: `np.ndarray`
        The energy spectra of the given hamiltonian
    """
    # Extract number of qubits
    n_qubits = hamiltonian.n_qubits

    # Define qubit register
    register = range(n_qubits)

    # Intialize energies
    energies = np.zeros((2 ** len(register)))

    # Obtain spectrum, scanning term by term
    for i, term in enumerate(hamiltonian.terms):
        # Extract coefficients
        out = np.real(hamiltonian.coeffs[i])

        # Compute tensor product
        for qubit in register:
            if qubit in term.qubit_indices:
                out = np.kron([1, -1], out)
            else:
                out = np.kron([1, 1], out)

        # Add energy contribution of the term
        energies += out

    # Add constant term to the spectrum
    energies = energies + hamiltonian.constant

    return energies


def plot_energy_spectrum(
    hamiltonian: Hamiltonian,
    high_k_states: Optional[int] = None,
    low_k_states: Optional[int] = None,
    ax=None,
    cmap="winter",
) -> None:
    """
    Compute and plot the energy spectrum of a given hamiltonian on
    a matplotlib figure.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object whose spectrum is computed.
    high_k_states: `int`, optional
        Optionally plot the highest k energy levels. Defaults to None.
    low_k_states: `int`, optional
        Optionally, plot the lowest k energy levels. Defaults to None.
    ax: Matplotlib axes object, optional
        Axes to plot on. Defaults to None.
    cmap: `str`, optional
        Specify the matplotlib colormap to use for plotting. Defaults to 'winter'.
    """
    # Compute energy spectrum
    energies = energy_spectrum_hamiltonian(hamiltonian)

    # Extract energy levels and their degeneracy
    unique_energies, degeneracy = np.unique(energies, return_counts=True)

    # If required extract highest or lowest k energy levels
    if high_k_states is not None:
        unique_energies, degeneracy = (
            unique_energies[-high_k_states:],
            degeneracy[-high_k_states:],
        )
    elif low_k_states is not None:
        unique_energies, degeneracy = (
            unique_energies[:low_k_states],
            degeneracy[:low_k_states],
        )

    # Define colormap
    cmap = plt.cm.get_cmap(cmap, len(unique_energies))

    # If no axis provided, define figure
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(3, 5))

    # Plot energy levels
    for i, energy in enumerate(unique_energies):
        ax.axhline(energy, label=f"Degeneracy={degeneracy[i]}", color=cmap(i))

    # Set axis attributes and legend
    ax.set(
        xticks=[],
        yticks=unique_energies,
        ylabel="Energy[a.u.]",
        title="Hamiltonian Energy spectrum",
    )
    ax.legend(loc="center left", fontsize=8)

    return None


def low_energy_states(
    hamiltonian: Hamiltonian, threshold_per: float
) -> Tuple[float, list]:
    """
    Return threshold energy and the low energy states of the
    specified hamiltonian which are below this threshold.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Compute the low energy states of this Hamiltonian
    threshold_per: `float`
        Threshold percentage away from the ground state,
        defining the energy we window we search in for low energy states.

    Returns
    -------
    low_energy_threshold: `float`
        The energy threshold below which we retrieve states.
    states: `list`
        The list of low energy states that lie below the low
        energy threshold.

    .. Important::
        The threshold is calculated as `threshols_per` factors away from the
        ground state of the Hamiltonian.
    """
    # Asserr threshold is bounded between 0 and 1
    assert threshold_per >= 0.0, "Threshold percentage should be above 0"
    assert threshold_per <= 1.0, "Threshold percentage should be below 1"

    # Compute energy spectrum of the Hamiltonian
    energies = energy_spectrum_hamiltonian(hamiltonian)

    # Extract ground state and highest excited state
    ground_state_energy = np.min(energies)
    highest_state_energy = np.max(energies)

    # Compute the low energy threshols
    low_energy_threshold = ground_state_energy + threshold_per * np.abs(
        highest_state_energy - ground_state_energy
    )

    # Initilize inndices for low energy states
    low_energy_indices = []

    # Obtain indices for energies below the threshold
    for idx in range(len(energies)):
        if energies[idx] <= low_energy_threshold:
            low_energy_indices.append(idx)

    # Extract states from the Hamiltonian spectrum
    states = [
        np.binary_repr(index, hamiltonian.n_qubits)[::-1]
        for index in low_energy_indices
    ]

    return low_energy_threshold, states


def low_energy_states_overlap(
    hamiltonian: Hamiltonian, threshold_per: float, prob_dict: dict
) -> float:
    """
    Calculates the overlap between the low energy states of a Hamiltonian,
    below a specific threshold away from the ground state energy, and an input
    state, expressed in terms of a probability dictionary.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Compute overlap with respect to energies of this Hamiltonian
    threshold_per: `float`
        Threshold percentage away from the ground state, defining the
        energy we window we search in for low energy states.
    prob_dict: `dict`
        The measurement outcome dictionary generated from the
        circuit execution.

    Returns
    -------
    total_overlap: `float`
        The total overlap with the low-energy states.

    .. Important::
        The threshold is calculated as `threshold_per` factors away
        from the ground state of the Hamiltonain.
        For `threshold_per=0` the function returns the ground state overlap of the QAOA output.
    """
    # Extract number of qubits from probability dictionary
    n_qubits = len(list(prob_dict.keys())[0])

    # Ensure number of qubits matches the number of qubits registered in the Hamiltonian
    assert (
        n_qubits == hamiltonian.n_qubits
    ), "Number of qubits in the Hamiltonian does not match the probabilities specified"

    # Extract low energy states
    _, states = low_energy_states(hamiltonian, threshold_per)

    # Compute overlap
    total_overlap = sum([prob_dict[state] for state in states]) / sum(
        list(prob_dict.values())
    )

    return total_overlap


def exp_val_single(spin: int, prob_dict: dict):
    """
    Computes expectation value <Z> of a given spin.

    Parameters
    ----------
    spin: `int`
        Spin whose expectation value we compute.
    prob_dict: `dict`
        Dictionary containing the configuration probabilities of each spin.

    Returns
    -------
    exp_val: `float`
        Expectation value of the spin
    """

    # Initialize expectation value
    exp_val = 0
    norm = sum(prob_dict.values())

    # Compute correlation
    for bitstring, prob in prob_dict.items():
        # If 0, spin is pointing up, else it is pointing down
        Z = int(bitstring[spin])

        # Add contribution if spin points up or subtract if points down
        exp_val += -prob / norm if Z > 0 else prob / norm

    return exp_val


def exp_val_pair(spins: tuple, prob_dict: dict):
    r"""
    Computes the correlation :math:`Mij = <Z_{i}Z_{j}>` between qubits i,j using the QAOA optimized
    wavefunction.

    .. Important::
        In the presence of linear terms the :math:`<Z_{i}><Z_{j}>` contribution needs to be
        subtracted later. This is done in the exp_val_hamiltonian_termwise() function used as a
        wrapper for this function.

    Parameters
    ----------
    spins: `tuple`
        Tuple containing the spins whose correlation is computed.
    prob_dict: `dict`
        The dictionary containing the configuration probabilities of each spin.

    Returns
    -------
    corr: `float`
        Correlation between the two spins in the term.

    """

    # Initialize correlation
    corr = 0

    norm = sum(prob_dict.values())
    # Compute correlation
    for bitstring, prob in prob_dict.items():
        # If 0 or 2, spins are aligned, else they are anti-aligned
        num_ones = sum([int(bitstring[i]) for i in spins])

        # Add contribution if spins aligned or subtract if anti-aligned
        corr += prob / norm if num_ones % 2 == 0 else -prob / norm

    return corr


def exp_val_hamiltonian_termwise(
    hamiltonian: Hamiltonian,
    mixer_type: str,
    p: int,
    qaoa_optimized_angles: Optional[list] = None,
    qaoa_optimized_counts: Optional[dict] = None,
    analytical: bool = True,
):
    """
    Computes the single spin expectation values <Z_{i}> and the correlation matrix Mij = <Z_{i}Z_{j}>,
    using the optimization results obtained from QAOA tranining the specified QAOA cost backend.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object containing the problem statement.
    p: `int`
        Number of layers in QAOA ansatz.
    qaoa_optimized_angles: `list`
        Optimized angles of the underlying QAOA.
    qaoa_optimized_counts: `dict`
        Dictionary containing the measurement counts of optimized QAOA circuit.
    analytical: `bool`
        Boolean that indicates whether to use analytical or numerical expectation
        calculation methods.

    Returns
    -------
    exp_vals_z: `np.array`
        Single spin expectation values as a numpy array.
    corr_matrix: `np.array`
        Correlation matrix as a numpy Matrix object.
    """

    # Define number of qubits, problem hamiltonian and QAOA parameters
    n_qubits = hamiltonian.n_qubits

    # Extract Hamiltonian terms
    terms = list(hamiltonian.terms)

    # Initialize the z expectation values and correlation matrix with 0s
    exp_vals_z = np.zeros(n_qubits)
    corr_matrix = np.zeros((n_qubits, n_qubits))

    # If single layer ansatz use analytical results
    if (
        analytical == True
        and p == 1
        and mixer_type == "x"
        and isinstance(qaoa_optimized_angles, list)
    ):
        # Compute expectation values and correlations of terms present in the Hamiltonian
        for term in terms:
            # If bias term compute expectation value
            if len(term) == 1:
                i = term.qubit_indices[0]
                exp_vals_z[i] = exp_val_single_analytical(
                    i, hamiltonian, qaoa_optimized_angles
                )

            # If two-body term compute correlation
            elif len(term) == 2:
                i, j = term.qubit_indices
                corr_matrix[i][j] = exp_val_pair_analytical(
                    (i, j), hamiltonian, qaoa_optimized_angles
                )
            else:
                continue

    # If multilayer ansatz, perform numerical computation
    else:
        if isinstance(qaoa_optimized_counts, dict):
            counts_dict = qaoa_optimized_counts
        else:
            raise ValueError(
                "Please specify optimized counts to compute expectation values."
            )

        # Compute expectation values and correlations of terms present in the Hamiltonian
        for term in terms:
            # If bias term compute expectation value
            if len(term) == 1:
                i = term.qubit_indices[0]
                exp_vals_z[i] = exp_val_single(i, counts_dict)

            # If two-body term compute correlation
            elif len(term) == 2:
                i, j = term.qubit_indices
                corr_matrix[i][j] = exp_val_pair((i, j), counts_dict)

            else:
                continue

    # Remove expectation value contribution from the correlations
    corr_matrix -= np.outer(exp_vals_z, exp_vals_z)

    return exp_vals_z, corr_matrix


@round_value
def calculate_calibration_factors(
    hamiltonian: Hamiltonian,
    calibration_measurements: dict,
    calibration_registers: list,
    qubit_mapping: list,
) -> Dict:
    """
    Computes the single spin and pairs of spins calibration factors, which are the expectation value of the observables found in the particular Hamiltonian, <Z_{i}> and <Z_{i}Z_{j}>, from the calibration data provided. The calibration data is obtained under BFA on an empty (initial state = |000..0>) QAOA circuit.
    See arXiv:2012.09738 and arXiv:2106.05800.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object containing the problem statement.
    calibration_measurements: `dict`
        Dictionary containing the measurement counts of an empty QAOA circuit.
    calibration_registers: `list`
        List specifying the physical (device) qubits on which the calibration data has been obtained.
        This is required because the calibration is usually performed over the whole device and hence the measurement outcomes (the calibration data) are strings with the size of the whole device while usually only a particular section is used.
    qubit_mapping: `list`
        List specifying the physical (device) qubits on which the QAOA circuit is executed. Related to qubit selection and qubit routing.

    Returns
    -------
    calibration_factors: `dict`
        Calibration factors as a dict.
    """
    calibration_registers_dict = {v: k for k, v in enumerate(calibration_registers)}

    # Define number of qubits, problem hamiltonian and QAOA parameters
    n_qubits = hamiltonian.n_qubits

    # Extract Hamiltonian terms
    terms = list(hamiltonian.terms)

    if qubit_mapping == None:
        qubit_mapping = np.arange(0, n_qubits)

    # Initialize an empty dict
    calibration_factors = {}

    # Compute single spin and pairs of spins expectation values of terms present in the Hamiltonian
    for term in terms:
        # If bias term compute single spins expectation value
        if len(term) == 1:
            i = term.qubit_indices[0]
            i_phys = qubit_mapping[i]
            i_cal = calibration_registers_dict[i_phys]
            exp_val_z = exp_val_single(i_cal, calibration_measurements)
            calibration_factors.update({(i,): exp_val_z})

        # If two-body term compute pairs of spins expectation values
        elif len(term) == 2:
            i, j = term.qubit_indices  # problem indices, ex: (0,1)
            i_phys, j_phys = (
                qubit_mapping[i],
                qubit_mapping[j],
            )  # physical indices, ex: (133, 131) after routing
            i_cal, j_cal = (
                calibration_registers_dict[i_phys],
                calibration_registers_dict[j_phys],
            )  # calibration indices, i.e. to which location on the measurement string each physical qubit corresponds to, ex: (63, 61)
            exp_val_zz = exp_val_pair((i_cal, j_cal), calibration_measurements)

            calibration_factors.update(
                {(i, j): exp_val_zz}
            )  # calibration factors are calculated for the terms in the hamiltonian/problem

        # If constant term, ignore
        if len(term) == 0:
            continue

    assert all(
        value != 0 for value in calibration_factors.values()
    ), "One (or more) of the calibration factors is 0 which means that the measurement is faulty. Please check the data."

    assert all(
        value <= 1 for value in calibration_factors.values()
    ), "One (or more) of the calibration factors is larger than 1 which is not physical. Please check the data."

    return calibration_factors


################################################################################
# ANALYTIC & KNOWN FORMULAE
################################################################################


def exp_val_single_analytical(spin: int, hamiltonian: Hamiltonian, qaoa_angles: tuple):
    """
    Computes the single spin expectation value :math:`<Z>` from an analytically
    derived expression for a single layer QAOA Ansatz.

    .. Important::
        Only valid for single layer QAOA Ansatz with X mixer Hamiltonian.

    Parameters
    ----------
    spin: `int`
        The spin whose expectation value we compute.

    hamiltonian: `Hamiltonian`
        Hamiltonian object containing the problem statement.

    qaoa_angles: `tuple`
        Pair of (gamma,beta) angles defined from QAOA ansatz and
        obtained in the QAOA process.

    Returns:
    -------
    exp_val: `float`
        Spin expectation value <Z>.
    """

    # Number of qubits in the system
    n_qubits = hamiltonian.n_qubits

    # Extract graph properties of the Hamiltonian
    terms = list(hamiltonian.terms)
    edges = [terms[j].qubit_indices for j in range(len(terms))]
    weights = hamiltonian.coeffs

    # Hamiltonian from graph definitions
    hamil_graph = dict(zip(edges, weights))

    # Spin biases
    h_u = hamil_graph[(spin,)] if hamil_graph.get((spin,)) is not None else 0

    # QAOA angles
    beta, gamma = qaoa_angles

    # Spin register as a list without the spin we focus on
    iter_qubits = [j for j in range(0, spin)] + [j for j in range(spin + 1, n_qubits)]

    # Initialize products
    exp_val = -np.sin(2 * beta) * np.sin(2 * gamma * h_u)

    # Loop over edges connecting u and v to other spins
    for n in iter_qubits:
        # Edges between the spin and others in the register
        edge = tuple([min(spin, n), max(spin, n)])

        # If edge not present in the graph the associated weight is set to 0
        J_un = 0 if hamil_graph.get(edge) is None else hamil_graph[edge]

        # Add factor to the products
        exp_val *= np.cos(2 * gamma * J_un)

    return exp_val


def exp_val_pair_analytical(spins: tuple, hamiltonian: Hamiltonian, qaoa_angles: tuple):
    """
    Computes :math:`<Z_{i}Z_{j}>` correlation between apair of spins analytically.
    It is an extension from the expression derived by
    Bravyi et al. in https://arxiv.org/abs/1910.08980 which includes the effect of biases.

    .. Important::
        * Only valid for single layer QAOA Ansatz with X mixer Hamiltonian.
        * In the presence of linear terms the <Z_{i}><Z_{j}> contribution needs to be subtracted later. This is done in the exp_val_hamiltonian_termwise() function used as a wrapper for this function.
        * OpenQAOA uses a different sign convention for the QAOA Ansatz than Bravy et al. - there is a relative minus sign between the cost function and the mixer in OpenQAOA, which is accounted for in this implementation. Additionally, the result below is valid for a Hadamard state initialization and in the absence of bias terms in the Hamiltonian.

    Parameters
    ----------
    spins: `tuple`
        Pair of spins whose correlation we compute.

    hamiltonian: `Hamiltonian`
        Hamiltonian object containing the problem statement.

    qaoa_angles: `tuple`
        Pair of (gamma,beta) angles defined from QAOA ansatz and
        obtained in the QAOA process.

    Returns
    -------
    corr: `float`
        Correlation <ZZ> between the specified spin pair.
    """

    # Number of qubits in the system
    n_qubits = hamiltonian.n_qubits

    # Extract graph properties of the Hamiltonian
    terms = list(hamiltonian.terms)
    edges = [terms[j].qubit_indices for j in range(len(terms))]
    weights = hamiltonian.coeffs

    # Hamiltonian from graph definitions
    hamil_graph = dict(zip(edges, weights))

    # Spins whose correlation we compute
    u, v = spins

    # Coupling between the spins
    J_uv = hamil_graph[spins] if hamil_graph.get(spins) is not None else 0

    # Spin biases
    h_u = hamil_graph[(u,)] if hamil_graph.get((u,)) is not None else 0
    h_v = hamil_graph[(v,)] if hamil_graph.get((v,)) is not None else 0

    # QAOA angles
    beta, gamma = qaoa_angles

    # Factors in the expression
    s = np.sin(2 * beta)
    c = np.cos(2 * beta)

    # Spin register as a list without u,v spins
    iter_qubits = (
        [j for j in range(0, min(u, v))]
        + [j for j in range(min(u, v) + 1, max(u, v))]
        + [j for j in range(max(u, v) + 1, n_qubits)]
    )

    # Initialize products
    prod1 = s**2 / 2 * np.cos(2 * gamma * (h_u - h_v))
    prod2 = -(s**2) / 2 * np.cos(2 * gamma * (h_u + h_v))
    prod3 = -c * s * np.sin(2 * gamma * J_uv) * np.cos(2 * gamma * h_u)
    prod4 = -c * s * np.sin(2 * gamma * J_uv) * np.cos(2 * gamma * h_v)

    # Loop over edges connecting u and v to other spins
    for n in iter_qubits:
        # Edges between u,v and another spin in the register
        edge1 = tuple([min(u, n), max(u, n)])
        edge2 = tuple([min(v, n), max(v, n)])

        # If edge not present in the graph the associated weight is set to 0
        J_un = 0 if hamil_graph.get(edge1) is None else hamil_graph[edge1]
        J_vn = 0 if hamil_graph.get(edge2) is None else hamil_graph[edge2]

        # Add factor to the products
        prod1 *= np.cos(2 * gamma * (J_un - J_vn))
        prod2 *= np.cos(2 * gamma * (J_un + J_vn))
        prod3 *= np.cos(2 * gamma * J_un)
        prod4 *= np.cos(2 * gamma * J_vn)

    # Add the contribution from each product term
    corr = prod1 + prod2 + prod3 + prod4

    return corr


def energy_expectation_analytical(angles: Union[list, tuple], hamiltonian: Hamiltonian):
    """
    Computes the expectation value of the Hamiltonian for an analytical expression.

    .. Important::
        Only valid for single layer QAOA Ansatz with X mixer Hamiltonian and classical
        Hamiltonians with up to quadratic terms.

    Parameters
    ----------
    angles: `list` or `tuple`
        QAOA angles at which the Hamiltonian expectation value is computed
    hamiltonian: `Hamiltonian`
        Classical Hamiltonian from which the expectation value is computed.
    """

    # Extract terms and coefficients from the Hamiltonian
    terms = [pauli_term.qubit_indices for pauli_term in hamiltonian.terms]
    coeffs = hamiltonian.coeffs

    energy = 0

    # Compute the expectation value of each term and add its local energy contribution
    for coeff, term in zip(coeffs, terms):
        if len(term) == 2:
            local_energy = exp_val_pair_analytical(term, hamiltonian, angles)

        else:
            local_energy = exp_val_single_analytical(term[0], hamiltonian, angles)

        energy += coeff * local_energy

    # Add constant shift contribution
    energy += hamiltonian.constant

    return energy


def ring_of_disagrees(reg: List[int]) -> Hamiltonian:
    """
    Builds the cost Hamiltonian for the "Ring of Disagrees".

    Parameters
    ----------
    reg: `list`
        register of qubits in the system.

    Returns
    -------
    ring_hamil: `Hamiltonian`
        Hamiltonian object containing Ring of Disagrees model.

    .. Important::
        This model is introduced in https://arxiv.org/abs/1411.4028
    """

    # Number of qubits from input register
    n_qubits = len(reg)

    # Define terms for the ring structure
    terms = [(reg[i], reg[(i + 1) % n_qubits]) for i in range(n_qubits)]

    # Define coefficients as in original formulation of the model
    coeffs = [0.5] * len(terms)

    # Constant term as in original formulation of the model
    constant = -len(terms) * 0.5

    # Define Hamiltonian
    ring_hamil = Hamiltonian.classical_hamiltonian(terms, coeffs, constant=constant)
    return ring_hamil


################################################################################
# OTHER MISCELLANEOUS
################################################################################


def flip_counts(counts_dictionary: dict) -> dict:
    """
    Returns a counts/probability dictionary that have their keys flipped. This
    formats the bit-strings from a right-most bit representing being the first
    qubit to the left-most bit representing the first qubit.

    Parameters
    ----------
    counts_dictionary: `dict`
        Count dictionary whose keys are flipped.

    Returns
    -------
    output_counts_dictionary: `dict`
        Count dictionary with flipped keys.
    """

    output_counts_dictionary = dict()

    for key, value in counts_dictionary.items():
        output_counts_dictionary[key[::-1]] = value

    return output_counts_dictionary


def permute_counts_dictionary(
    counts_dictionary: dict, final_qubit_layout: List[int]
) -> dict:
    """Permutes the order of the qubits in the counts dictionary to the
    original order if SWAP gates were used leading to modified qubit layout.
    Parameters
    ----------
    counts_dictionary : `dict`
        The measurement outcomes obtained from the Simulator/QPU
    original_qubit_layout: List[int]
        The qubit layout in which the qubits were initially
    final_qubit_layout: List[int]
        The final qubit layout after application of SWAPs

    Returns
    -------
    `dict`
        The permuted counts dictionary with qubits in the original place
    """

    # Create a mapping of original positions to final positions
    original_qubit_layout = list(range(len(final_qubit_layout)))
    mapping = {
        original_qubit_layout[i]: final_qubit_layout[i]
        for i in range(len(original_qubit_layout))
    }
    permuted_counts = {}

    for basis, counts in counts_dictionary.items():

        def permute_string(basis_state: str = basis, mapping: dict = mapping):
            # Use the mapping to permute the string
            permuted_string = "".join(
                [basis_state[mapping[i]] for i in range(len(basis_state))]
            )
            return permuted_string

        permuted_counts.update({permuted_string: counts})

    return permuted_counts


def negate_counts_dictionary(counts_dictionary: dict, s: int) -> dict:
    """Negates every bitstring of the counts dictionary according to
    the position of the X gates before the measurement.
    Used in SPAM Twirling.
    Parameters
    ----------
    counts_dictionary : `dict`
        The measurement outcomes obtained from the Simulator/QPU
    s: int
        Syndrome whose binary representation denotes the negated qubits. For example, 4 = 100, signifies that the first qubit had an X gate just before the measurement, which requires the first digit of the every key to be classically negated inside this function.

    Returns
    -------
    `dict`
        The negated counts dictionary
    """
    negated_counts = {}
    for key in counts_dictionary.keys():
        n_qubits = len(key)
        negated_key = s ^ int(
            key, 2
        )  # bitwise XOR to classically negate randomly chosen qubits, specified by s
        negated_counts.update(
            [(format(negated_key, "b").zfill(n_qubits), counts_dictionary[key])]
        )
    return negated_counts


@round_value
def qaoa_probabilities(statevector) -> dict:
    """
    Return a qiskit-style probability dictionary from a statevector.

    Parameters
    ----------
    statevector: `np.ndarray[complex]`
        The wavefunction whose probability distribution needs to be calculated.

    Returns
    -------
    prob_dict: `dict`
        Probabilities represented as a python dictionary with basis states stored
        as keys and their probabilities as their corresponding values.
    """
    # Define list of probabilities from wavefunction amplitudes
    prob_vec = np.real(np.conjugate(statevector) * statevector)

    # Extract number of qubits from size of probability
    n_qubits = int(np.log2(len(prob_vec)))

    # Initialize probability dictionary
    prob_dict = {}

    for x in range(len(prob_vec)):
        # Define binary representation of each state, with qubit-0 most significant bit
        key = np.binary_repr(x, n_qubits)[::-1]

        # Update probability dictionary
        prob_dict.update({key: prob_vec[x]})

    return prob_dict


################################################################################
# DICTIONARY MANIPULATION and SERIALIZATION
################################################################################
def delete_keys_from_dict(obj: Union[list, dict], keys_to_delete: List[str]):
    """
    Recursively delete all the keys keys_to_delete from a object (or list of dictionaries)
    Parameters
    ----------
    obj: dict or list[dict]
        dictionary or list of dictionaries from which we want to delete keys
    keys_to_delete: list
        list of keys to delete from the dictionaries

    Returns
    -------
    obj: dict or list[dict]
        dictionary or list of dictionaries from which we have deleted keys
    """
    if isinstance(obj, dict):
        for key in keys_to_delete:
            if key in obj:
                del obj[key]
        for key in obj:
            if isinstance(obj[key], dict):
                delete_keys_from_dict(obj[key], keys_to_delete)
            elif isinstance(obj[key], list):
                for item in obj[key]:
                    delete_keys_from_dict(item, keys_to_delete)
    elif isinstance(obj, list):
        for item in obj:
            delete_keys_from_dict(item, keys_to_delete)

    return obj


def convert2serialize(obj, complex_to_string: bool = False):
    """
    Recursively converts object to dictionary.

    Parameters
    ----------
    obj: object
        Object to convert to dictionary.
    complex_to_string: bool
        If True, convert complex numbers to string, so the result can be serialized to JSON.

    Returns
    -------
    dict: dict
        Dictionary representation of the object.
    """
    if isinstance(obj, dict):
        return {
            k: convert2serialize(v, complex_to_string)
            for k, v in obj.items()
            if v is not None
        }
    elif hasattr(obj, "_ast"):
        return convert2serialize(obj._ast(), complex_to_string)
    elif isinstance(obj, tuple):
        return tuple(
            convert2serialize(v, complex_to_string) for v in obj if v is not None
        )
    elif not isinstance(obj, str) and hasattr(obj, "__iter__"):
        return [convert2serialize(v, complex_to_string) for v in obj if v is not None]
    elif hasattr(obj, "__dict__"):
        return {
            k: convert2serialize(v, complex_to_string)
            for k, v in obj.__dict__.items()
            if not callable(v) and v is not None
        }
    elif complex_to_string and isinstance(obj, complex):
        return str(obj)
    else:
        return obj


################################################################################
# UUID and Timestamp
################################################################################


def generate_timestamp() -> str:
    """
    Generate a timestamp string in UTC+0. Format: YYYY-MM-DDTHH:MM:SS.

    Returns
    -------
    timestamp: `str`
        String representation of a timestamp.
    """
    return datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S")


def generate_uuid() -> str:
    """
    Generate a UUID string.

    Returns
    -------
    uuid: `str`
        String representation of a UUID.
    """
    return str(uuid.uuid4())


def is_valid_uuid(uuid_to_test: str) -> bool:
    """
    Check if a string is a valid UUID.

    Parameters
    ----------
    uuid_to_test: `str`
        String to check if it is a valid UUID.

    Returns
    -------
    is_valid: `bool`
        Boolean indicating if the string is a valid UUID.
    """

    try:
        # generate a UUID object from the string, if it is a valid UUID it won't throw an error
        _ = uuid.UUID(uuid_to_test, version=4)
        return True
    except ValueError:
        # If it's a value error, then the string is not a valid string for a UUID.
        return False


def permute_counts_dictionary(
    counts_dictionary: dict, permutation_order: List[int]
) -> dict:
    """Permutes the order of the qubits in the counts dictionary to the
    original order if SWAP gates were used leading to modified qubit layout.
    Parameters
    ----------
    counts_dictionary : `dict`
        The measurement outcomes obtained from the Simulator/QPU
    permutation_order: List[int]
        The qubit order to permute the dictionary with

    Returns
    -------
    `dict`
        The permuted counts dictionary with qubits in the original place
    """

    # Create a mapping of original positions to final positions
    # original order always goes from 0 -> n-1
    original_order = list(range(len(permutation_order)))
    mapping = {
        original_order[i]: permutation_order[i] for i in range(len(original_order))
    }
    permuted_counts = {}

    for basis_state, counts in counts_dictionary.items():
        # Use the mapping to permute the string
        permuted_string = "".join(
            [basis_state[mapping[i]] for i in range(len(basis_state))]
        )
        permuted_counts.update({permuted_string: counts})

    return permuted_counts


################################################################################
# CHECKING FUNCTION
################################################################################


def check_kwargs(list_expected_params, list_default_values, **kwargs):
    """
    Checks that the given list of expected parameters can be found in the
    kwargs given as input. If so, it returns the parameters from kwargs, else
    it raises an exception.

    Args:
        list_expected_params: List[str]
            List of string containing the name of the expected parameters in
            kwargs
        list_default_values: List
            List containing the deafult values of the expected parameters in
            kwargs
        **kwargs:
            Keyword arguments where keys are supposed to be the expected params

    Returns:
        A tuple with the actual expected parameters if they are found in kwargs.

    Raises:
        ValueError:
            If one of the expected arguments is not found in kwargs and its
            default value is not specified.
    """

    def check_kwarg(expected_param, default_value, **kwargs):
        param = kwargs.pop(expected_param, default_value)

        if param is None:
            raise ValueError(f"Parameter '{expected_param}' should be specified")

        return param

    params = []
    for expected_param, default_value in zip(list_expected_params, list_default_values):
        params.append(check_kwarg(expected_param, default_value, **kwargs))

    return tuple(params)


####################################################################################
# QAOALIB
####################################################################################


def dicke_basis(excitations: int, n_qubits: int) -> np.ndarray:
    """
    Generates the Dicke basis state $|ek>$ with $k$ excitations

    Parameters
    ----------
    excitations: `int`
        Number of excitations in the basis vector
    n_qubits: `int`
        Total number of qubits in the system

    Returns
    -------
    `np.ndarray`
        Total basis states present in the expected Dicke vector in the computational basis
    """
    assert (
        n_qubits >= excitations
    ), "Excitations cannot be larger than total qubits in system"
    sub_sys_excitations = np.ones(excitations, dtype=int)
    sub_sys_ground = np.zeros(n_qubits - excitations, dtype=int)

    total_state = np.concatenate((sub_sys_ground, sub_sys_excitations))
    total_basis_comp = set(itertools.permutations(total_state))
    total_basis_comp = np.array(
        ["".join(str(i) for i in basis_comp) for basis_comp in total_basis_comp]
    )

    return total_basis_comp


def dicke_wavefunction(excitations, n_qubits):
    """
    Generate the k-excitations dicke statevector

    Parameters
    ----------
    excitations: int
        The number of excitations in the basis

    n_qubits: int
        The number of qubits in the register

    Returns
    -------
    `np.ndarray`
        The wavefunction vector for a given cumulative Dicke states with $<=k$ excitations
    """

    k_dicke = dicke_basis(excitations, n_qubits)
    k_dicke_ints = [int(state, 2) for state in k_dicke]
    wavefunction = np.array(
        [
            1.0 + 0.0j if num in k_dicke_ints else 0.0 + 0.0j
            for num in range(2**n_qubits)
        ],
        dtype=complex,
    ) / np.sqrt(len(k_dicke_ints))
    return wavefunction


def k_cumulative_excitations(k: int, n_qubits: int) -> np.ndarray:
    """
    Generates the Upper bound excitations basis vector $|Ek>$, which a superposition of all
    Dicke basis vectors upto excitation number $k$

    Parameters
    ----------
    k: `int`
        Upper bound on number of excitations in the basis vector
    n_qubits: `int`
        Total number of qubits in the system

    Returns
    -------
    wavefunction: `np.ndarray`
        The wavefunction vector for a given cumulative Dicke states with $<=k$ excitations
    """
    cumulative_dicke_bases = np.array(["0" * n_qubits])
    for exc in range(1, k + 1):
        cumulative_dicke_bases = np.concatenate(
            (cumulative_dicke_bases, dicke_basis(exc, n_qubits))
        )

    wavefn_locs = [int(basis, 2) for basis in cumulative_dicke_bases]
    wavefunction = np.array(
        [1 if loc in wavefn_locs else 0 for loc in range(2**n_qubits)], dtype=complex
    ) / np.sqrt(len(wavefn_locs))

    return wavefunction


def knapsack_balanced_basis(
    weight_capacity: int,
    weights_list: List,
    decision_register: List,
    slack_register: List,
):
    """
    Generates the basis where the system register is balanced for Knapsack Hamiltonian, i.e. Slack register
    compensating the decision register state to cancel the penalty term in Hamiltonian.
    NOTE: Cannot cancel the penalty term if decision register weight exceeds the weight capacity of the sack

    Parameters
    ----------
    weight_capacity:
        The knapsack capacity, i.e. the upper limit on the weights that can be added in the knapsack
    weights_list:
        List of item weights in the problem
    decision_register:
        The qubit regsiter of the decision bits
    slack_register:
        The qubit regsiter of the slack bits

    Returns
    -------
    np.ndarray
    """
    n_decision_qubits = len(decision_register)
    n_slack_qubits = len(slack_register)
    n_total_qubits = n_slack_qubits + n_decision_qubits

    def to_bin(number, n_qubits):
        # not using np.binary_repr because it is deprecated!
        binary_form = bin(number)[2:].zfill(n_qubits)
        return binary_form

    decision_config_weights = {
        to_bin(dec_i, n_decision_qubits): sum(
            [
                weight * int(to_bin(dec_i, n_decision_qubits)[i])
                for i, weight in enumerate(weights_list)
            ]
        )
        for dec_i in range(2**n_decision_qubits)
    }

    decision_slack_configs = {
        to_bin(dec_i, n_decision_qubits): (
            to_bin(
                weight_capacity
                - decision_config_weights[to_bin(dec_i, n_decision_qubits)],
                n_slack_qubits,
            )
            if decision_config_weights[to_bin(dec_i, n_decision_qubits)]
            < weight_capacity
            else to_bin(0, n_slack_qubits)
        )
        for dec_i in range(2**n_decision_qubits)
    }

    all_configs = []
    for dec_config, slack_config in decision_slack_configs.items():
        config = np.empty(n_total_qubits, dtype=str)
        for i, loc in enumerate(decision_register):
            config[loc] = dec_config[i]
        for i, loc in enumerate(slack_register[::-1]):
            config[loc] = slack_config[i]

        config_str = "".join(i for i in config)
        all_configs.append(config_str[::-1])

    wavefn_locs = [int(basis, 2) for basis in all_configs]
    wavefunction = np.array(
        [1 if loc in wavefn_locs else 0 for loc in range(2**n_total_qubits)],
        dtype=complex,
    ) / np.sqrt(len(wavefn_locs))

    return wavefunction



import unittest
from unittest.mock import Mock
import json
import numpy as np
import pytest
import subprocess

import openqaoa
from openqaoa_azure.backends import DeviceAzure
from openqaoa.qaoa_components import (
    create_qaoa_variational_params,
    PauliOp,
    Hamiltonian,
    QAOADescriptor,
    QAOAVariationalStandardParams,
)
from openqaoa_qiskit.backends import (
    QAOAQiskitQPUBackend,
)
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.problems import NumberPartition
from openqaoa import QAOA


class TestingQAOAQiskitQPUBackendAzure(unittest.TestCase):
    
    """This Object tests the QAOA Qiskit QPU Backend object with the Azure 
    Device object. This checks that the use of qiskit to send circuits to Azure
    is working as intended.
    
    The Azure CLI has to be configured beforehand to run these tests.
    """
    
    @pytest.mark.api
    def setUp(self):
        
        bashCommand = "az resource list"
        process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
        output, error = process.communicate()
        
        if error is not None:
            print(error)
            raise Exception('You must have the Azure CLI installed and must be logged in to use the Azure Quantum Backends')
        else:
            output_json = json.loads(output)
            output_json_s = [each_json for each_json in output_json if each_json['name'] == 'TestingOpenQAOA'][0]
            self.RESOURCE_ID = output_json_s['id']
            self.AZ_LOCATION = output_json_s['location']
            
    @pytest.mark.sim
    def check_shots_tally(self):
        
        """There is a known bug in the qiskit backend for azure where if the shots 
        argument might be ignored. This test checks that the output from the azure 
        computation matches the input requirements.
        """
        
        shots = 1024
        problem_qubo = NumberPartition([1, 2, 3]).qubo
        azure_device = create_device(location='azure', name='rigetti.sim.qvm', 
                                     resource_id=self.RESOURCE_ID, 
                                     az_location=self.AZ_LOCATION)
        
        q = QAOA()
        q.set_device(azure_device)
        q.set_backend_properties(n_shots=shots)
        q.set_classical_optimizer(maxiter = 1)
        q.compile(problem_qubo)
        q.optimize()
        
        comp_shots = sum(q.result.optimized['optimized measurement outcomes'].values())
        
        self.assertEqual(shots, comp_shots)
        
    @pytest.mark.api
    def test_expectations_in_init(self):
        
        """
        Testing the Exceptions in the init function of the QiskitQPUShotBasedBackend
        """
        
        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [1/8*np.pi]
        betas = [1/8*np.pi]
        shots = 10000

        cost_hamil = Hamiltonian([PauliOp('ZZ', (0, 1)), PauliOp('ZZ', (1, 2)),
                                  PauliOp('ZZ', (0, 2))], weights, 1)
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor,
                                                       betas, gammas)
        
        # We mock the potential Exception that could occur in the Device class
        azure_device = DeviceAzure('', '', '')
        azure_device._check_provider_connection = Mock(return_value=False)
        
        try:
            QAOAQiskitQPUBackend(qaoa_descriptor, azure_device, 
                                 shots, None, None, True)
        except Exception as e:
            self.assertEqual(str(e), 'Error connecting to AZURE.')
        
        with self.assertRaises(Exception):
            QAOAQiskitQPUBackend(qaoa_descriptor, azure_device, shots, None, None, True)
        

        azure_device = DeviceAzure(device_name='', resource_id=self.RESOURCE_ID, 
                                   az_location=self.AZ_LOCATION)
        
        try:
            QAOAQiskitQPUBackend(qaoa_descriptor, azure_device, 
                                 shots, None, None, True)
        except Exception as e:
            self.assertEqual(str(e), 'Connection to AZURE was made. Error connecting to the specified backend.')
        
        with self.assertRaises(Exception):
            QAOAQiskitQPUBackend(qaoa_descriptor, azure_device, shots, None, None, True)
        
    @pytest.mark.api
    def test_remote_qubit_overflow(self):
        
        """
        If the user creates a circuit that is larger than the maximum circuit size
        that is supported by the QPU. An Exception should be raised with the 
        appropriate error message alerting the user to the error.
        """
        
        shots = 100
        
        set_of_numbers = np.random.randint(1, 10, 6).tolist()
        qubo = NumberPartition(set_of_numbers).qubo

        mixer_hamil = X_mixer_hamiltonian(n_qubits=6)
        qaoa_descriptor = QAOADescriptor(qubo.hamiltonian, mixer_hamil, p=1)
        variate_params = create_qaoa_variational_params(qaoa_descriptor, 'standard', 'rand')

        azure_device = DeviceAzure('rigetti.sim.qvm', self.RESOURCE_ID, self.AZ_LOCATION)
        
        try:
            QAOAQiskitQPUBackend(qaoa_descriptor, azure_device, 
                                 shots, None, None, True)
        except Exception as e:
            self.assertEqual(str(e), 'There are lesser qubits on the device than the number of qubits required for the circuit.')


if __name__ == "__main__":
    unittest.main()



# def stochastic_grad_descent(
#     fun,
#     x0,
#     jac,
#     args=(),
#     stepsize=0.001,
#     mass=0.9,
#     startiter=0,
#     maxiter=1000,
#     callback=None,
#     **options
# ):
#     """
#     Scipy "OptimizeResult" object for method == 'sgd'
#     scipy.optimize.minimize compatible implementation of stochastic gradient descent with momentum.
#     Adapted from ``autograd/misc/optimizers.py``.

#     """
#     x = x0
#     velocity = np.zeros_like(x)
#     print("SGD Test")

#     for i in range(startiter, startiter + maxiter):
#         g = jac(x)

#         velocity = mass * velocity - (1.0 - mass) * g
#         x = x + stepsize * velocity

#         if callback is not None:
#             callback(x)

#     return OptimizeResult(x=x, fun=fun(x), jac=g, nit=i + 1, nfev=i + 1, success=True)



import networkx as nx
from docplex.mp.model import Model
from typing import Union
import numpy as np
import itertools
import copy
import matplotlib.pyplot as plt
from matplotlib import colormaps

from .problem import Problem
from .converters import FromDocplex2IsingModel
from .qubo import QUBO


class VRP(Problem):
    """
    Creates an instance of the vehicle routing problem (VRP).
    https://en.wikipedia.org/wiki/Vehicle_routing_problem

    Parameters
    ----------
    G: networkx.Graph
        Networkx graph of the problem
    pos: list[list]
        position x, y of each node
    n_vehicles: int
        the number of vehicles used in the solution
    depot: int
        the node where all the vehicles leave for and return after
    subtours: list[list]
        if -1 (Default value): All the possible subtours are added to the constraints. Avoid it for large instances.
        if there are subtours that want be avoided in the solution, e.g, a 8 nodes
        VRP with an optimal solution showing subtour between nodes 4, 5, and 8 can be
        avoided introducing the constraint subtours=[[4,5,8]]. To additional information
        about subtours refer to https://de.wikipedia.org/wiki/Datei:TSP_short_cycles.png
    all_subtours:
    penalty: float
        Penalty for the constraints. If the method is 'unbalanced' three values are needed,
        one for the equality constraints and two for the inequality constraints.
    method: str
        Two available methods for the inequality constraints ["slack", "unbalanced"]
        For 'unblanced' see https://arxiv.org/abs/2211.13914
    Returns
    -------
        An instance of the VRP problem.
    """

    __name__ = "vehicle_routing"

    def __init__(
        self,
        G,
        pos,
        n_vehicles,
        depot: int = 0,
        subtours: list = -1,
        method: str = "slack",
        penalty: Union[int, float, list] = 4,
    ):
        self.G = G
        if len(G.nodes) != len(pos):
            raise ValueError(
                f"The number of nodes in G is {len(G.nodes)} while the x, y coordinates in pos is {len(pos)}"
            )
        self.pos = pos
        self.n_vehicles = n_vehicles
        self.depot = depot
        self.subtours = subtours
        self.method = method
        if method == "unbalanced" and len(penalty) != 3:
            raise ValueError(
                "The penalty must have 3 parameters [lambda_0, lambda_1, lambda_2]"
            )
        else:
            self.penalty = penalty

    @property
    def G(self):
        return self._G

    @G.setter
    def G(self, input_networkx_graph):
        if not isinstance(input_networkx_graph, nx.Graph):
            raise TypeError("Input problem graph must be a networkx Graph.")

        # Relabel nodes to integers starting from 0
        mapping = dict(
            zip(input_networkx_graph, range(input_networkx_graph.number_of_nodes()))
        )
        self._G = nx.relabel_nodes(input_networkx_graph, mapping)

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the vehicle routing problem, whose graph is
        random.

        Parameters
        ----------
        **kwargs:
            Required keyword arguments are:
            n_nodes: int
                The number of nodes (vertices) in the graph.
            n_vehicles: int
                Number of vehicles used in the problem
            method: str
                method for the inequality constraints ['slack', 'unbalanced'].
                For the unbalaced method refer to https://arxiv.org/abs/2211.13914
                For the slack method: https://en.wikipedia.org/wiki/Slack_variable
            subtours: list[list]
                Manually add the subtours to be avoided
            seed: int
                Seed for the random problems.

        Returns
        -------
            A random instance of the vehicle routing problem.
        """
        n_nodes = kwargs.get("n_nodes", 6)
        n_vehicles = kwargs.get("n_vehicles", 2)
        seed = kwargs.get("seed", None)
        method = kwargs.get("method", "slack")
        if method == "slack":
            penalty = kwargs.get("penalty", 4)
        elif method == "unbalanced":
            penalty = kwargs.get("penalty", [4, 1, 1])
        else:
            raise ValueError(f"The method '{method}' is not valid.")
        subtours = kwargs.get("subtours", -1)
        np.random.seed(seed)
        G = nx.Graph()
        G.add_nodes_from(range(n_nodes))
        pos = [[0, 0]]
        pos += [list(2 * np.random.rand(2) - 1) for _ in range(n_nodes - 1)]
        for i in range(n_nodes - 1):
            for j in range(i + 1, n_nodes):
                r = np.sqrt((pos[i][0] - pos[j][0]) ** 2 + (pos[i][1] - pos[j][1]) ** 2)
                G.add_weighted_edges_from([(i, j, r)])
        return VRP(
            G, pos, n_vehicles, subtours=subtours, method=method, penalty=penalty
        )

    @property
    def docplex_model(self):
        """
        Return a docplex model of the vehicle routing problem using the linear
        programming approach. In this approach, the edges between different nodes
        of the graph are the variables of the problem. Therefore, x_i_j will represent
        if the path between nodes i and j is selected. The idea is to minimize the distance
        traveled

        sum_{i, j > i} r_{ij} x_{ij}.

        where r_{ij} is the distance between nodes i and j (self.G.edges[i, j]["weight"]).

        Returns
        -------
        mdl : Model
            Docplex model of the Vehicle routing problem.

        """
        mdl = Model("VRP")
        num_nodes = self.G.number_of_nodes()
        # Variables: the edges between nodes for a symmetric problem. This means x_i_j == x_j_i and only x_i_j is used.
        x = {
            (i, j): mdl.binary_var(name=f"x_{i}_{j}")
            for i in range(num_nodes - 1)
            for j in range(i + 1, num_nodes)
        }
        # Distance traveled
        mdl.minimize(
            mdl.sum(
                self.G.edges[i, j]["weight"] * x[(i, j)]
                for i in range(num_nodes - 1)
                for j in range(i + 1, num_nodes)
            )
        )
        # Only 2 edges for node constraint
        for i in range(num_nodes):
            if i != self.depot:
                mdl.add_constraint(
                    mdl.sum(
                        [x[tuple(sorted([i, j]))] for j in range(num_nodes) if i != j]
                    )
                    == 2
                )
            else:
                mdl.add_constraint(
                    mdl.sum(
                        [x[tuple(sorted([i, j]))] for j in range(num_nodes) if i != j]
                    )
                    == 2 * self.n_vehicles
                )

        if self.subtours == -1:
            # When subtours are not added, by default all the possible subtours are added.
            # However, it is really costly for large instances (above 10 nodes)
            list_subtours = [[i for i in range(num_nodes) if i != self.depot]]
            for nodes in list_subtours:  # costly for large instances
                for i in range(3, num_nodes - 2 * self.n_vehicles):
                    for subtour in itertools.combinations(nodes, i):
                        tour = sorted(subtour)
                        n_subtour = len(subtour)
                        mdl.add_constraint(
                            mdl.sum(
                                [
                                    x[(tour[i], tour[j])]
                                    for i in range(n_subtour - 1)
                                    for j in range(i + 1, n_subtour)
                                ]
                            )
                            <= n_subtour - 1
                        )
        # Subtour constraints if any
        elif isinstance(self.subtours, list):
            list_subtours = self.subtours
            for subtour in list_subtours:
                tour = sorted(subtour)
                n_subtour = len(subtour)
                if n_subtour != 0:
                    mdl.add_constraint(
                        mdl.sum(
                            [
                                x[(tour[i], tour[j])]
                                for i in range(n_subtour)
                                for j in range(i + 1, n_subtour)
                            ]
                        )
                        <= n_subtour - 1
                    )
        else:
            raise ValueError(
                f"{type(self.subtour)} is not a valid format for the subtours."
            )
        return mdl

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
            The QUBO encoding of this problem.
        """
        cplex_model = self.docplex_model
        if self.method == "slack":
            qubo_docplex = FromDocplex2IsingModel(
                cplex_model, multipliers=self.penalty
            ).ising_model
        elif self.method == "unbalanced":
            qubo_docplex = FromDocplex2IsingModel(
                cplex_model,
                multipliers=self.penalty[0],
                unbalanced_const=True,
                strength_ineq=self.penalty[1:],
            ).ising_model
        return QUBO(
            qubo_docplex.n,
            qubo_docplex.terms + [[]],
            qubo_docplex.weights + [qubo_docplex.constant],
            self.problem_instance,
        )

    def classical_solution(self, string: bool = False):
        """
        Return the classical solution of the vehicle routing problem
        Parameters
        ----------
        string : bool, optional
            If the solution is returned as a string. The default is False.
        Raises
        ------
        ValueError
            A flag if docplex does not find a valid solution.
        Returns
        -------
        solution : Union[str, dict]
            The classical solution of the specific problem as a string or a dict.
        """
        cplex_model = self.docplex_model
        docplex_sol = cplex_model.solve()

        if docplex_sol is None:
            raise ValueError(f"Solution not found: {cplex_model.solve_details.status}.")

        if string:
            solution = ""
        else:
            solution = {}
        for var in cplex_model.iter_binary_vars():
            if string:
                solution += str(round(docplex_sol.get_value(var)))
            else:
                solution[var.name] = round(docplex_sol.get_value(var))
        return solution

    def paths_subtours(self, sol):
        """
        Return the routes the different vehicles take and the subtours the solution
        has if any
        Parameters
        ----------
        sol : dict
            Solution to the vrp with x_i_j the edge between nodes i and j.

        Returns
        -------
        solution: dict
        Dictionary with two keys:
        - 'paths': solutions of the VRP,
        - 'subtours': subtours found in the solution
        """
        n_nodes = self.G.number_of_nodes()
        vars_list = []
        for i in range(n_nodes - 1):
            for j in range(i + 1, n_nodes):
                if round(sol[f"x_{i}_{j}"]):
                    vars_list.append([i, j])
        # ----------------  vehicle routing problem solutions -----------------
        paths = {}
        for i in range(self.n_vehicles):
            paths[i] = []
            nodes = [self.depot]
            count_depot = 0
            max_edges = n_nodes * (n_nodes - 1) / 2
            while count_depot < 2:
                for edge in vars_list:
                    if nodes[-1] in edge:
                        if self.depot in edge:
                            count_depot += 1
                        paths[i].append(edge)
                        vars_list.remove(edge)
                        nodes.append(edge[0] if edge[0] != nodes[-1] else edge[1])
                        break
                max_edges -= 1
                if max_edges < 0:
                    raise ValueError(
                        "Solution provided does not fulfill all the path conditions."
                    )
        # ----------------            subtours                -----------------
        subtours = {}
        i = 0
        max_edges = n_nodes * (n_nodes - 1) / 2
        while len(vars_list) > 0:
            subtours[i] = [vars_list.pop(0)]
            nodes = copy.copy(subtours[i][0])
            count = 1
            while count < 2:
                for edge in vars_list:
                    if nodes[-1] in edge:
                        if nodes[0] in edge:
                            count += 1
                        subtours[i].append(edge)
                        vars_list.remove(edge)
                        nodes.append(edge[0] if edge[0] != nodes[-1] else edge[1])
                        break
                max_edges -= 1
                if max_edges < 0:
                    raise ValueError("The subtours in the solution are broken.")
            i += 1
        return {"paths": paths, "subtours": subtours}

    def plot_solution(
        self, solution: Union[dict, str], ax=None, edge_width=4, colors=None
    ):
        """
        A visualization method for the vehicle routing problem solution.
        Parameters
        ----------
        solution : Union[dict, str]
            The solution of the specific vehicle routing problem as a string or dictionary.
        ax : matplotlib axes, optional
            The default is None.
        Returns
        -------
        fig : matplotlib.pyplot.Figure()
            The graph visualization of the solution.
        """

        colors = colormaps["jet"] if colors is None else colors
        if type(colors) is list and len(colors) != self.n_vehicles:
            raise ValueError(
                f"The length of colors {len(colors)} and the number of vehicles {self.n_vehicles} do not match"
            )

        if isinstance(solution, str):
            sol = {}
            for n, var in enumerate(self.docplex_model.iter_binary_vars()):
                sol[var.name] = int(solution[n])
            solution = sol
        paths_and_subtours = self.paths_subtours(solution)
        paths = paths_and_subtours["paths"]
        subtours = paths_and_subtours["subtours"]
        tours_color = {}
        for vehicle in range(self.n_vehicles):
            for i, j in paths[vehicle]:
                if type(colors) is list:
                    tours_color[f"x_{i}_{j}"] = colors[vehicle]
                else:
                    tours_color[f"x_{i}_{j}"] = colors((vehicle + 1) / self.n_vehicles)
        for subtour in subtours.keys():
            for i, j in subtours[subtour]:
                tours_color[f"x_{i}_{j}"] = "black"
        color_node = "#5EB1EB"
        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = None
        pos = self.pos
        num_vertices = self.G.number_of_nodes()
        G = nx.Graph()
        G.add_nodes_from(range(num_vertices))
        edge_color = []
        for i in range(num_vertices - 1):
            for j in range(i + 1, num_vertices):
                if int(solution[f"x_{i}_{j}"]):
                    edge_color.append(tours_color[f"x_{i}_{j}"])
                    G.add_edge(i, j)
        nx.draw(
            G,
            pos=pos,
            width=edge_width,
            edge_color=edge_color,
            node_color=color_node,
            alpha=0.8,
            labels={i: str(i) for i in range(num_vertices)},
            ax=ax,
            edgecolors="black",
        )
        return fig



import unittest
import numpy as np

from openqaoa.backends.qaoa_backend import get_qaoa_backend
from openqaoa.qaoa_components import (
    QAOADescriptor,
    QAOAVariationalStandardParams,
    Hamiltonian,
)
from openqaoa.utilities import random_classical_hamiltonian, X_mixer_hamiltonian
from openqaoa.backends import DeviceLocal


class TestGetSamplesMethod(unittest.TestCase):
    """
    Test the `get_samples` method in `cost_function.py`
    -> Test whether it produces the right results while sampling the wavefunction
    """

    def test_samples_with_one_string(self):
        """
        This function generates samples for a wavefunction that has
        unit probability for one a basis state and zero for others.
        In other words, the final wavefunction is a product state
        in the computational basis.
        """
        # 6-qubit hamiltonian with only linear terms and no quadratic terms
        reg = [0, 1, 2, 3, 4, 5]
        solution = bin(0)[2:].zfill(len(reg))
        terms = [(i,) for i in reg]
        weights = [1] * len(terms)
        p = 1
        betas, gammas = [np.pi / 4], [-np.pi / 4]

        cost_hamiltonian = Hamiltonian.classical_hamiltonian(terms, weights, constant=0)
        mixer_hamiltonian = X_mixer_hamiltonian(len(reg))
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = QAOAVariationalStandardParams(
            qaoa_descriptor, betas, gammas
        )

        backend_vectorized = get_qaoa_backend(
            qaoa_descriptor, DeviceLocal("vectorized")
        )

        shot_results_vec = backend_vectorized.sample_from_wavefunction(
            variational_params_std, n_samples=15
        )

        bool_list_vec = [x == solution for x in shot_results_vec]

        assert all(bool_list_vec)

    def test_samples_limiting_case(self):
        """
        Check whether sample distribution approximates the probability vector in the limit of
        large number of shots

        NOTE: if the assertion error still keeps failing for no apparent reason, try changing
              ``decimal`` argument in `np.testing.assert_array_almost_equal` to something lower!
        """
        # testing a large number of shots!
        nshots = 500000

        # random 6-qubit Hamiltonian with betas, gammas = 0,0 => only hadamards
        reg = [0, 1, 2, 3, 4, 5]
        p = 1

        betas, gammas = ([0], [0])

        cost_hamiltonian = random_classical_hamiltonian(reg)
        n_qubits = cost_hamiltonian.n_qubits
        mixer_hamiltonian = X_mixer_hamiltonian(n_qubits)
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = QAOAVariationalStandardParams(
            qaoa_descriptor, betas, gammas
        )

        backend_vectorized = get_qaoa_backend(
            qaoa_descriptor, DeviceLocal("vectorized")
        )

        # wf_vec = backend_vectorized.wavefunction(variational_params_std)
        prob_wf_vec = np.array(
            list(backend_vectorized.probability_dict(variational_params_std).values()),
            dtype=float,
        )

        samples_vec = backend_vectorized.sample_from_wavefunction(
            variational_params_std, n_samples=nshots
        )
        samples_dict_vec = {bin(x)[2:].zfill(n_qubits): 0 for x in range(2**n_qubits)}


        for shot_result in samples_vec:
            samples_dict_vec[shot_result] += 1 / nshots

        samples_prob_vec = np.array(list(samples_dict_vec.values()), dtype=float)

        np.testing.assert_array_almost_equal(prob_wf_vec, samples_prob_vec, decimal=3)

    # def testing_w_init_prog(self):
    # 	"""
    # 	This function creates an init_prog and then implements a Hamiltonian
    # 	with zero betas and gammas.
    # 	"""
    # 	reg = [0,1,2,3,4,5]
    # 	init_terms = [(i,) for i in reg]
    # 	init_weights = [1]*len(init_terms)
    # 	init_p=2

    # 	init_params = ([[np.pi/4]*len(reg),
    # 					[np.pi/2,0,0,0,0,np.pi/2]],
    # 					[[-np.pi/4]*len(init_terms),
    # 					[0]*len(reg)],
    # 					[[],[]])
    # 	hyperparams_init = HyperParams(init_terms,init_weights,reg,init_p)
    # 	init_prog = ExtendedParams(hyperparams_init,init_params)

    # 	terms,weights = zip(*random_hamiltonian(reg).items())
    # 	terms, weights = list(terms),list(weights)
    # 	p=1
    # 	trainable_params = ([0],[0])
    # 	hyperparams = HyperParams(terms,weights,reg,p)

    # 	qaoa_params = StandardParams(hyperparams,trainable_params)
    # 	cost_fn = QAOACostVector(qaoa_params,init_prog=init_prog)

    # 	solution = '100001'
    # 	shot_results = cost_fn.get_samples(qaoa_params.raw(), nshots=15)
    # 	bool_list = [x == solution for x in shot_results]

    # 	assert all(bool_list)


if __name__ == "__main__":
    unittest.main()



"""
Different hyper-parameterisations and trainable-parameterisations for QAOA

Hyper-Parameters help with problem hamiltonian initialisation and fixing `p`
for the QAOA problem

AbstractParams and child classes let one choose the parameterisation for
solving the QAOA problem in hand.  
"""

from .ansatz_constructor import *
from .variational_parameters import *



from setuptools import setup, find_namespace_packages
import os

current_path = os.getcwd()

with open("README.md", "r") as fh:
    long_description = fh.read()

# Dev package will share versions with the core in it.
with open("_version.py") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")

requirements = [f"{each_folder_name}=={version}" for each_folder_name in os.listdir("src") if "openqaoa-" in each_folder_name]
    
setup(
    name="openqaoa",
    python_requires=">=3.8, <3.11",
    version=version,
    author="Entropica Labs",
    entry_points={"openqaoa.plugins": []},
    url="https://github.com/entropicalabs/openqaoa",
    install_requires=requirements,
    license="MIT",
    description="OpenQAOA is a python open-source multi-backend Software Development Kit to create, customise and execute the Quantum Approximate Optimisation Algorithm (QAOA) on Noisy Intermediate-Scale Quantum (NISQ) devices, and simulators",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    keywords="quantum optimisation SDK",
)



from qiskit_ibm_provider import IBMProvider
from qiskit_aer import AerSimulator
from typing import List
from openqaoa.backends.basedevice import DeviceBase


class DeviceQiskit(DeviceBase):
    """
    Contains the required information and methods needed to access remote
    qiskit QPUs.

    Attributes
    ----------
    available_qpus: `list`
      When connection to a provider is established, this attribute contains a list
      of backend names which can be used to access the selected backend by reinitialising
      the Access Object with the name of the available backend as input to the
      device_name parameter.
    n_qubits: `int`
        The maximum number of qubits available for the selected backend. Only
        available if check_connection method is executed and a connection to the
        qpu and provider is established.
    """

    def __init__(
        self,
        device_name: str,
        hub: str = None,
        group: str = None,
        project: str = None,
        as_emulator: bool = False,
    ):
        """The user's IBMQ account has to be authenticated through qiskit in
        order to use this backend. This can be done through `IBMQ.save_account`.

        See: https://quantum-computing.ibm.com/lab/docs/iql/manage/account/ibmq

        Parameters
        ----------
                device_name: `str`
                        The name of the IBMQ device to be used
        hub: `str`
            Valid IBMQ hub name.
        group: `str`
            Valid IBMQ group name.
        project: `str`
            The name of the project for which the experimental data will be
            saved in on IBMQ's end.
        """

        self.device_name = device_name
        self.device_location = "ibmq"
        self.hub = hub
        self.group = group
        self.project = project
        self.as_emulator = as_emulator

        self.provider_connected = None
        self.qpu_connected = None

    def check_connection(self) -> bool:
        """
        This method should allow a user to easily check if the credentials
        provided to access the remote QPU is valid.

        If no backend was specified in initialisation of object, just runs
        a test connection without a specific backend.
        If backend was specified, checks if connection to that backend
        can be established.

        Returns
        -------
        bool
                        True if successfully connected to IBMQ or IBMQ and the QPU backend
                        if it was specified. False if unable to connect to IBMQ or failure
                        in the attempt to connect to the specified backend.
        """

        self.provider_connected = self._check_provider_connection()

        if self.provider_connected == False:
            return self.provider_connected

        self.available_qpus = [backend.name for backend in self.provider.backends()]

        if self.device_name == "":
            return self.provider_connected

        self.qpu_connected = self._check_backend_connection()

        if self.provider_connected and self.qpu_connected:
            return True
        else:
            return False

    def _check_backend_connection(self) -> bool:
        """Private method for checking connection with backend(s)."""

        if self.device_name in self.available_qpus:
            self.backend_device = self.provider.get_backend(self.device_name)
            self.n_qubits = self.backend_device.configuration().n_qubits
            if self.as_emulator is True:
                self.backend_device = AerSimulator.from_backend(self.backend_device)
            return True
        else:
            print(f"Please choose from {self.available_qpus} for this provider")
            return False

    def _check_provider_connection(self) -> bool:
        """
        Private method for checking connection with provider.
        """

        try:
            # Use default
            self.provider = IBMProvider()
            # Unless exact instance is specified
            if all([self.hub, self.group, self.project]):
                instance_name = self.hub + "/" + self.group + "/" + self.project
                assert instance_name in self.provider.instances()
                self.provider = IBMProvider(instance=instance_name)
            elif any([self.hub, self.group, self.project]):
                # if only partially specified, print warning.
                raise Exception(
                    "You've only partially specified the instance name. Either"
                    "the hub, group or project is missing. hub: {}, group: {}, project: {}.\n"
                    "The default instance will be used instead. (This default can "
                    "be specified when doing `IBMProvider.save_account`)"
                )
            return True
        except Exception as e:
            print(
                "An Exception has occured when trying to connect with the provider."
                "Please note that you are required to set up your IBMQ account locally first."
                "See: https://quantum-computing.ibm.com/lab/docs/iql/manage/account/ibmq"
                "for how to save your IBMQ account locally. \n {}".format(e)
            )
            return False

    def connectivity(self) -> List[List[int]]:
        return self.backend_device.configuration().coupling_map
